from math import pi

import numpy as np
from ase.units import Hartree

from gpaw.xc.sic import SIC
from gpaw.atom.generator import Generator
from gpaw.atom.configurations import parameters
from gpaw.utilities import hartree


class NSCFSIC:
    def __init__(self, paw):
        self.paw = paw

    def calculate(self):
        ESIC = 0
        xc = self.paw.hamiltonian.xc
        assert xc.type == 'LDA'

        # Calculate the contribution from the core orbitals
        for a in self.paw.density.D_asp:
            setup = self.paw.density.setups[a]
            # TODO: Use XC which has been used to calculate the actual
            # calculation.
            # TODO: Loop over setups, not atoms.
            print('Atom core SIC for ', setup.symbol)
            print('%10s%10s%10s' % ('E_xc[n_i]', 'E_Ha[n_i]', 'E_SIC'))
            g = Generator(setup.symbol, xcname='LDA', nofiles=True, txt=None)
            g.run(**parameters[setup.symbol])
            njcore = g.njcore
            for f, l, e, u in zip(g.f_j[:njcore], g.l_j[:njcore],
                                  g.e_j[:njcore], g.u_j[:njcore]):
                # Calculate orbital density
                # NOTE: It's spherically symmetrized!
                # n = np.dot(self.f_j,
                assert l == 0, ('Not tested for l>0 core states')
                na = np.where(abs(u) < 1e-160, 0, u)**2 / (4 * pi)
                na[1:] /= g.r[1:]**2
                na[0] = na[1]
                nb = np.zeros(g.N)
                v_sg = np.zeros((2, g.N))
                vHr = np.zeros(g.N)
                Exc = xc.calculate_spherical(g.rgd, np.array([na, nb]), v_sg)
                hartree(0, na * g.r * g.dr, g.r, vHr)
                EHa = 2 * pi * np.dot(vHr * na * g.r, g.dr)
                print(('%10.2f%10.2f%10.2f' % (Exc * Hartree, EHa * Hartree,
                                               -f * (EHa + Exc) * Hartree)))
                ESIC += -f * (EHa + Exc)

        sic = SIC(finegrid=True, coulomb_factor=1, xc_factor=1)
        sic.initialize(self.paw.density, self.paw.hamiltonian, self.paw.wfs)
        sic.set_positions(self.paw.spos_ac)

        print('Valence electron sic ')
        print('%10s%10s%10s%10s%10s%10s' % ('spin', 'k-point', 'band',
                                            'E_xc[n_i]', 'E_Ha[n_i]', 'E_SIC'))
        assert len(self.paw.wfs.kpt_u) == 1, 'Not tested for bulk calculations'

        for s, spin in sic.spin_s.items():
            spin.initialize_orbitals()
            spin.update_optimal_states()
            spin.update_potentials()

            n = 0
            for xc, c in zip(spin.exc_m, spin.ecoulomb_m):
                print(('%10i%10i%10i%10.2f%10.2f%10.2f' %
                       (s, 0, n, -xc * Hartree, -c * Hartree,
                        2 * (xc + c) * Hartree)))
                n += 1

            ESIC += spin.esic

        print('Total correction for self-interaction energy:')
        print('%10.2f eV' % (ESIC * Hartree))
        print('New total energy:')
        total = (ESIC * Hartree + self.paw.get_potential_energy() +
                 self.paw.get_reference_energy())
        print('%10.2f eV' % total)
        return total



import warnings
from math import sqrt, pi
import numpy as np

from ase.units import Ha
from gpaw import BadParallelization
from gpaw.mpi import world
from gpaw.density import redistribute_array, redistribute_atomic_matrices
from gpaw.sphere.lebedev import weight_n
from gpaw.utilities import pack, pack_atomic_matrices, unpack_atomic_matrices
from gpaw.xc.gllb import safe_sqr
from gpaw.xc.gllb.contribution import Contribution

# XXX Work in process
debug = False


def d(*args):
    if debug:
        print(args)


class ResponsePotential:
    """Container for response potential"""
    def __init__(self, response, vt_sG, vt_sg, D_asp, Dresp_asp):
        self.response = response
        self.vt_sG = vt_sG
        self.vt_sg = vt_sg
        self.D_asp = D_asp
        self.Dresp_asp = Dresp_asp

    def redistribute(self, new_response):
        old_response = self.response
        new_wfs = new_response.wfs
        new_density = new_response.density
        self.vt_sG = redistribute_array(self.vt_sG,
                                        old_response.density.gd,
                                        new_density.gd,
                                        new_wfs.nspins,
                                        new_wfs.kptband_comm)
        if self.vt_sg is not None:
            self.vt_sg = redistribute_array(self.vt_sg,
                                            old_response.density.finegd,
                                            new_density.finegd,
                                            new_wfs.nspins,
                                            new_wfs.kptband_comm)

        def redist_asp(D_asp):
            return redistribute_atomic_matrices(D_asp,
                                                new_density.gd,
                                                new_wfs.nspins,
                                                new_density.setups,
                                                new_density.atom_partition,
                                                new_wfs.kptband_comm)

        self.D_asp = redist_asp(self.D_asp)
        self.Dresp_asp = redist_asp(self.Dresp_asp)
        self.response = new_response


class C_Response(Contribution):
    """Response contribution for GLLB functionals.

    Parameters
    ----------
    weight
        Weight of the contribution
    coefficients
        Coefficient calculator object
    metallic
        If True, then Fermi level is used as the reference
        energy for coefficients instead of the HOMO energy.
        This is necessary to get sensible results in metallic systems.
    damp
        Small value to damp divisions by zero
    """
    def __init__(self,
                 weight: float,
                 coefficients, *,
                 metallic: bool = False,
                 damp: float = 1e-10):
        Contribution.__init__(self, weight)
        d('In c_Response __init__', self)
        self.coefficients = coefficients
        self.vt_sg = None
        self.vt_sG = None
        self.D_asp = None
        self.Dresp_asp = None
        self.Ddist_asp = None
        self.Drespdist_asp = None
        self.damp = damp
        self.fix_potential = False
        self.metallic = metallic

        # For logging reference energy
        self.eref_s = None
        self.eref_source_s = None

    def set_damp(self, damp):
        self.damp = damp

    def get_name(self):
        return 'RESPONSE'

    def get_desc(self):
        desc = []
        if self.metallic:
            desc += ['metallic']
        desc += [self.coefficients.get_description()]
        return ', '.join(desc)

    def initialize(self, density, hamiltonian, wfs):
        Contribution.initialize(self, density, hamiltonian, wfs)
        self.coefficients.initialize(wfs)
        if self.Dresp_asp is None:
            assert self.density.D_asp is None
        # XXX With the deprecated `fixdensity=True` option this
        # initialize() function is called both before AND after read()!
        # Thus, the second call of initialize() would discard the read
        # potential unless we check if it was already allocated.
        # Remove this if statement once `fixdensity=True` option has
        # been removed from the calculator.
        if self.vt_sg is None:
            self.vt_sG = self.gd.empty(self.nspins)
            self.vt_sg = self.finegd.empty(self.nspins)

    def initialize_1d(self, ae):
        Contribution.initialize_1d(self, ae)
        self.coefficients.initialize_1d(ae)

    def initialize_from_other_response(self, response):
        pot = ResponsePotential(response,
                                response.vt_sG,
                                response.vt_sg,
                                response.D_asp,
                                response.Dresp_asp)
        pot.redistribute(self)
        self.vt_sG = pot.vt_sG
        self.vt_sg = pot.vt_sg
        self.D_asp = pot.D_asp
        self.Dresp_asp = pot.Dresp_asp
        self.Ddist_asp = self.distribute_D_asp(pot.D_asp)
        self.Drespdist_asp = self.distribute_D_asp(pot.Dresp_asp)
        self.eref_s = response.eref_s
        self.eref_source_s = response.eref_source_s

    # Calcualte the GLLB potential and energy 1d
    def add_xc_potential_and_energy_1d(self, v_g):
        w_i = self.coefficients.get_coefficients_1d()
        u2_j = safe_sqr(self.ae.u_j)
        v_g += self.weight * np.dot(w_i, u2_j) / (
            np.dot(self.ae.f_j, u2_j) + self.damp)
        return 0.0  # Response part does not contribute to energy

    def update_potentials(self):
        d('In update response potential')
        if self.fix_potential:
            # Skip the evaluation of the potential so that
            # the existing potential is used. This is relevant for
            # band structure calculation so that the potential
            # does not get updated with the other k-point sampling.
            return

        if self.wfs.kpt_u[0].eps_n is None:
            # This skips the evaluation of the potential so that
            # the existing one is used.
            # This happens typically after reading when occupations
            # haven't been calculated yet and the potential read earlier
            # is used.
            return

        # Calculate reference energy
        self.eref_s = []
        self.eref_source_s = []
        if self.metallic:
            # Use Fermi level as reference levels
            fermilevel = self.wfs.fermi_level
            assert isinstance(fermilevel, float), \
                'GLLBSCM supports only a single Fermi level'
            for s in range(self.wfs.nspins):
                self.eref_source_s.append('Fermi level')
                self.eref_s.append(fermilevel)
        else:
            # Find homo and lumo levels for each spin
            for s in range(self.wfs.nspins):
                homo, lumo = self.wfs.get_homo_lumo(s)
                # Check that homo and lumo are reasonable
                if homo > lumo:
                    # HOMO higher than LUMO; set Fermi level as reference
                    fermilevel = self.wfs.fermi_level
                    self.eref_source_s.append('Fermi level')
                    self.eref_s.append(fermilevel)
                else:
                    self.eref_source_s.append('HOMO')
                    self.eref_s.append(homo)

        w_kn = self.coefficients.get_coefficients(self.wfs.kpt_u,
                                                  eref_s=self.eref_s)
        f_kn = [kpt.f_n for kpt in self.wfs.kpt_u]
        if w_kn is not None:
            self.vt_sG[:] = 0.0
            nt_sG = self.gd.zeros(self.nspins)

            for kpt, w_n in zip(self.wfs.kpt_u, w_kn):
                self.wfs.add_to_density_from_k_point_with_occupation(
                    self.vt_sG, kpt, w_n)
                self.wfs.add_to_density_from_k_point(nt_sG, kpt)

            self.wfs.kptband_comm.sum(nt_sG)
            self.wfs.kptband_comm.sum(self.vt_sG)

            if self.wfs.kd.symmetry:
                for nt_G, vt_G in zip(nt_sG, self.vt_sG):
                    self.wfs.kd.symmetry.symmetrize(nt_G, self.gd)
                    self.wfs.kd.symmetry.symmetrize(vt_G, self.gd)

            d('response update D_asp', world.rank, self.Dresp_asp.keys(),
              self.D_asp.keys())
            self.wfs.calculate_atomic_density_matrices_with_occupation(
                self.Dresp_asp, w_kn)
            self.Drespdist_asp = self.distribute_D_asp(self.Dresp_asp)
            d('response update Drespdist_asp', world.rank,
              self.Dresp_asp.keys(), self.D_asp.keys())
            self.wfs.calculate_atomic_density_matrices_with_occupation(
                self.D_asp, f_kn)
            self.Ddist_asp = self.distribute_D_asp(self.D_asp)

            self.vt_sG /= nt_sG + self.damp

        self.density.distribute_and_interpolate(self.vt_sG, self.vt_sg)

    def calculate(self, e_g, n_sg, v_sg):
        self.update_potentials()
        v_sg += self.weight * self.vt_sg

    def distribute_D_asp(self, D_asp):
        return self.density.atomdist.to_work(D_asp)

    def calculate_energy_and_derivatives(self, setup, D_sp, H_sp, a,
                                         addcoredensity=True):
        # Get the XC-correction instance
        c = setup.xc_correction
        ncresp_g = setup.extra_xc_data['core_response'] / self.nspins
        if not addcoredensity:
            ncresp_g[:] = 0.0

        for D_p, dEdD_p, Dresp_p in zip(self.Ddist_asp[a], H_sp,
                                        self.Drespdist_asp[a]):
            D_Lq = np.dot(c.B_pqL.T, D_p)
            n_Lg = np.dot(D_Lq, c.n_qg)  # Construct density
            if addcoredensity:
                n_Lg[0] += c.nc_g * sqrt(4 * pi) / self.nspins
            nt_Lg = np.dot(
                D_Lq, c.nt_qg
            )  # Construct smooth density (_without smooth core density_)

            Dresp_Lq = np.dot(c.B_pqL.T, Dresp_p)
            nresp_Lg = np.dot(Dresp_Lq, c.n_qg)  # Construct 'response density'
            nrespt_Lg = np.dot(
                Dresp_Lq, c.nt_qg
            )  # Construct smooth 'response density' (w/o smooth core)

            for w, Y_L in zip(weight_n, c.Y_nL):
                nt_g = np.dot(Y_L, nt_Lg)
                nrespt_g = np.dot(Y_L, nrespt_Lg)
                x_g = nrespt_g / (nt_g + self.damp)
                dEdD_p -= self.weight * w * np.dot(
                    np.dot(c.B_pqL, Y_L), np.dot(c.nt_qg, x_g * c.rgd.dv_g))

                n_g = np.dot(Y_L, n_Lg)
                nresp_g = np.dot(Y_L, nresp_Lg)
                x_g = (nresp_g + ncresp_g) / (n_g + self.damp)

                dEdD_p += self.weight * w * np.dot(
                    np.dot(c.B_pqL, Y_L), np.dot(c.n_qg, x_g * c.rgd.dv_g))

        return 0.0

    def integrate_sphere(self, a, Dresp_sp, D_sp, Dwf_p, spin=0):
        c = self.wfs.setups[a].xc_correction
        Dresp_p, D_p = Dresp_sp[spin], D_sp[spin]
        D_Lq = np.dot(c.B_pqL.T, D_p)
        n_Lg = np.dot(D_Lq, c.n_qg)  # Construct density
        n_Lg[0] += c.nc_g * sqrt(4 * pi) / len(D_sp)
        nt_Lg = np.dot(D_Lq, c.nt_qg
                       )  # Construct smooth density (without smooth core)
        Dresp_Lq = np.dot(c.B_pqL.T, Dresp_p)  # Construct response
        nresp_Lg = np.dot(Dresp_Lq, c.n_qg)  # Construct 'response density'
        nrespt_Lg = np.dot(
            Dresp_Lq, c.nt_qg
        )  # Construct smooth 'response density' (w/o smooth core)
        Dwf_Lq = np.dot(c.B_pqL.T, Dwf_p)  # Construct lumo wf
        nwf_Lg = np.dot(Dwf_Lq, c.n_qg)
        nwft_Lg = np.dot(Dwf_Lq, c.nt_qg)
        E = 0.0
        for w, Y_L in zip(weight_n, c.Y_nL):
            v = np.dot(Y_L, nwft_Lg) * np.dot(Y_L, nrespt_Lg) / (
                np.dot(Y_L, nt_Lg) + self.damp)
            E -= self.weight * w * np.dot(v, c.rgd.dv_g)
            v = np.dot(Y_L, nwf_Lg) * np.dot(Y_L, nresp_Lg) / (
                np.dot(Y_L, n_Lg) + self.damp)
            E += self.weight * w * np.dot(v, c.rgd.dv_g)
        return E

    def add_smooth_xc_potential_and_energy_1d(self, vt_g):
        w_ln = self.coefficients.get_coefficients_1d(smooth=True)
        v_g = np.zeros(self.ae.N)
        n_g = np.zeros(self.ae.N)
        for w_n, f_n, u_n in zip(w_ln, self.ae.f_ln,
                                 self.ae.s_ln):  # For each angular momentum
            u2_n = safe_sqr(u_n)
            v_g += np.dot(w_n, u2_n)
            n_g += np.dot(f_n, u2_n)

        vt_g += self.weight * v_g / (n_g + self.damp)
        return 0.0  # Response part does not contribute to energy

    def calculate_delta_xc(self, homolumo=None):
        warnings.warn(
            'The function calculate_delta_xc() is deprecated. '
            'Please use calculate_discontinuity_potential() instead. '
            'See documentation on calculating band gap with GLLBSC.',
            DeprecationWarning)

        if homolumo is None:
            # Calculate band gap

            # This always happens, so we don't warn.
            # We should perhaps print it as an ordinary message,
            # but we do not have a file here to which to print.
            # import warnings
            # warnings.warn('Calculating KS-gap directly from the k-points, '
            #               'can be inaccurate.')
            homolumo = self.wfs.get_homo_lumo()
        homo, lumo = homolumo
        dxc_pot = self.calculate_discontinuity_potential(homo * Ha, lumo * Ha)
        self.Dxc_vt_sG = dxc_pot.vt_sG
        self.Dxc_D_asp = dxc_pot.D_asp
        self.Dxc_Dresp_asp = dxc_pot.Dresp_asp

    def calculate_discontinuity_potential(self, homo, lumo):
        r"""Calculate the derivative discontinuity potential.

        This function calculates $`\Delta_{xc}(r)`$ as given by
        Eq. (24) of https://doi.org/10.1103/PhysRevB.82.115106

        Parameters:

        homo:
            homo energy (or energies in spin-polarized case) in eV
        lumo:
            lumo energy (or energies in spin-polarized case) in eV

        Returns:

        dxc_pot: Discontinuity potential
        """
        if self.nspins == 2:
            eref_s = np.asarray(homo)
            eref_lumo_s = np.asarray(lumo)
        else:
            eref_s = np.asarray([homo])
            eref_lumo_s = np.asarray([lumo])
        assert len(eref_s) == len(eref_lumo_s) == self.nspins
        eref_s /= Ha
        eref_lumo_s /= Ha

        dxc_Dresp_asp = self.empty_atomic_matrix()
        dxc_D_asp = self.empty_atomic_matrix()
        for a in self.density.D_asp:
            ni = self.wfs.setups[a].ni
            dxc_Dresp_asp[a] = np.zeros((self.nspins, ni * (ni + 1) // 2))
            dxc_D_asp[a] = np.zeros((self.nspins, ni * (ni + 1) // 2))

        # Calculate new response potential with LUMO reference
        w_kn = self.coefficients.get_coefficients_for_lumo_perturbation(
            self.wfs.kpt_u, eref_s=eref_s, eref_lumo_s=eref_lumo_s)
        f_kn = [kpt.f_n for kpt in self.wfs.kpt_u]

        dxc_vt_sG = self.gd.zeros(self.nspins)
        nt_sG = self.gd.zeros(self.nspins)
        for kpt, w_n in zip(self.wfs.kpt_u, w_kn):
            self.wfs.add_to_density_from_k_point_with_occupation(dxc_vt_sG,
                                                                 kpt, w_n)
            self.wfs.add_to_density_from_k_point(nt_sG, kpt)

        self.wfs.kptband_comm.sum(nt_sG)
        self.wfs.kptband_comm.sum(dxc_vt_sG)

        if self.wfs.kd.symmetry:
            for nt_G, dxc_vt_G in zip(nt_sG, dxc_vt_sG):
                self.wfs.kd.symmetry.symmetrize(nt_G, self.gd)
                self.wfs.kd.symmetry.symmetrize(dxc_vt_G, self.gd)

        dxc_vt_sG /= nt_sG + self.damp

        self.wfs.calculate_atomic_density_matrices_with_occupation(
            dxc_Dresp_asp, w_kn)
        self.wfs.calculate_atomic_density_matrices_with_occupation(
            dxc_D_asp, f_kn)
        dxc_pot = ResponsePotential(self, dxc_vt_sG, None,
                                    dxc_D_asp, dxc_Dresp_asp)
        return dxc_pot

    def calculate_delta_xc_perturbation(self):
        warnings.warn(
            'The function calculate_delta_xc_perturbation() is deprecated. '
            'Please use calculate_discontinuity() instead. '
            'See documentation on calculating band gap with GLLBSC.',
            DeprecationWarning)
        dxc_pot = ResponsePotential(self, self.Dxc_vt_sG, None, self.Dxc_D_asp,
                                    self.Dxc_Dresp_asp)
        if self.nspins != 1:
            ret = []
            for spin in range(self.nspins):
                ret.append(self.calculate_discontinuity(dxc_pot, spin=spin))
            return ret
        return self.calculate_discontinuity(dxc_pot)

    def calculate_delta_xc_perturbation_spin(self, s=0):
        warnings.warn(
            'The function calculate_delta_xc_perturbation_spin() is '
            'deprecated. Please use calculate_discontinuity_spin() instead. '
            'See documentation on calculating band gap with GLLBSC.',
            DeprecationWarning)
        dxc_pot = ResponsePotential(self, self.Dxc_vt_sG, None, self.Dxc_D_asp,
                                    self.Dxc_Dresp_asp)
        return self.calculate_discontinuity(dxc_pot, spin=s)

    def calculate_discontinuity(self,
                                dxc_pot: ResponsePotential,
                                spin: int = None):
        r"""Calculate the derivative discontinuity.

        This function evaluates the perturbation theory expression
        for the derivative discontinuity value as given by
        Eq. (25) of https://doi.org/10.1103/PhysRevB.82.115106

        Parameters:

        dxc_pot:
            Discontinuity potential.
        spin:
            Spin value.

        Returns:

        KSgap:
            Kohn-Sham gap in eV
        dxc:
            Derivative discontinuity in eV
        """
        if spin is None:
            if self.nspins != 1:
                raise ValueError('Specify spin for the discontinuity.')
            spin = 0

        # Redistribute discontinuity potential
        if dxc_pot.response is not self:
            dxc_pot.redistribute(self)

        homo, lumo = self.wfs.get_homo_lumo(spin)
        KSgap = lumo - homo

        # Find the lumo-orbital of this spin
        if self.wfs.bd.comm.size != 1:
            raise BadParallelization(
                'Band parallelization is not supported by '
                'calculate_discontinuity().')
        eps_n = self.wfs.kpt_qs[0][spin].eps_n
        lumo_n = (eps_n < self.wfs.fermi_level).sum()

        # Calculate the expectation value for all k points, and keep
        # the smallest energy value
        nt_G = self.gd.empty()
        min_energy = np.inf
        for u, kpt in enumerate(self.wfs.kpt_u):
            if kpt.s == spin:
                nt_G[:] = 0.0
                self.wfs.add_orbital_density(nt_G, kpt, lumo_n)
                E = 0.0
                for a in dxc_pot.D_asp:
                    D_sp = dxc_pot.D_asp[a]
                    Dresp_sp = dxc_pot.Dresp_asp[a]
                    P_ni = kpt.P_ani[a]
                    Dwf_p = pack(np.outer(P_ni[lumo_n].T.conj(),
                                          P_ni[lumo_n]).real)
                    E += self.integrate_sphere(a, Dresp_sp, D_sp, Dwf_p,
                                               spin=spin)
                E = self.gd.comm.sum(E)
                E += self.gd.integrate(nt_G * dxc_pot.vt_sG[spin])
                E += kpt.eps_n[lumo_n] - lumo
                min_energy = min(min_energy, E)

        # Take the smallest value over all distributed k points
        dxc = self.wfs.kd.comm.min(min_energy)
        return KSgap * Ha, dxc * Ha

    def calculate_discontinuity_from_average(self,
                                             dxc_pot: ResponsePotential,
                                             spin: int,
                                             testing: bool = False):
        msg = ('This function estimates discontinuity by calculating '
               'the average of the discontinuity potential. '
               'Use only for testing and debugging.')
        if not testing:
            raise RuntimeError(msg)
        else:
            warnings.warn(msg)
        assert self.wfs.world.size == 1

        # Calculate average of lumo reference response potential
        dxc = np.average(dxc_pot.vt_sG[spin])
        return dxc * Ha

    def initialize_from_atomic_orbitals(self, basis_functions):
        # Initialize 'response-density' and density-matrices
        # print('Initializing from atomic orbitals')
        self.Dresp_asp = self.empty_atomic_matrix()
        setups = self.wfs.setups

        for a in self.density.D_asp.keys():
            ni = setups[a].ni
            self.Dresp_asp[a] = np.zeros((self.nspins, ni * (ni + 1) // 2))

        self.D_asp = self.empty_atomic_matrix()
        f_asi = {}
        w_asi = {}

        for a in basis_functions.atom_indices:
            if setups[a].type == 'ghost':
                w_j = [0.0]
            else:
                w_j = setups[a].extra_xc_data['w_j']

            # Basis function coefficients based of response weights
            w_si = setups[a].calculate_initial_occupation_numbers(
                0, False,
                charge=0,
                nspins=self.nspins,
                f_j=w_j)
            # Basis function coefficients based on density
            f_si = setups[a].calculate_initial_occupation_numbers(
                0, False,
                charge=0,
                nspins=self.nspins)

            if a in basis_functions.my_atom_indices:
                self.Dresp_asp[a] = setups[a].initialize_density_matrix(w_si)
                self.D_asp[a] = setups[a].initialize_density_matrix(f_si)

            f_asi[a] = f_si
            w_asi[a] = w_si

        self.Drespdist_asp = self.distribute_D_asp(self.Dresp_asp)
        self.Ddist_asp = self.distribute_D_asp(self.D_asp)
        nt_sG = self.gd.zeros(self.nspins)
        basis_functions.add_to_density(nt_sG, f_asi)
        self.vt_sG = self.gd.zeros(self.nspins)
        basis_functions.add_to_density(self.vt_sG, w_asi)
        # Update vt_sG to correspond atomic response potential. This will be
        # used until occupations and eigenvalues are available.
        self.vt_sG /= nt_sG + self.damp
        self.vt_sg = self.finegd.zeros(self.nspins)
        self.density.distribute_and_interpolate(self.vt_sG, self.vt_sg)

    def get_extra_setup_data(self, extra_data):
        ae = self.ae
        njcore = ae.njcore
        w_ln = self.coefficients.get_coefficients_1d(smooth=True)
        w_j = []
        for w_n in w_ln:
            for w in w_n:
                w_j.append(w)
        extra_data['w_j'] = w_j

        w_j = self.coefficients.get_coefficients_1d()
        x_g = np.dot(w_j[:njcore], safe_sqr(ae.u_j[:njcore]))
        x_g[1:] /= ae.r[1:] ** 2 * 4 * np.pi
        x_g[0] = x_g[1]
        extra_data['core_response'] = x_g

        # For debugging purposes
        w_j = self.coefficients.get_coefficients_1d()
        u2_j = safe_sqr(self.ae.u_j)
        v_g = self.weight * np.dot(w_j, u2_j) / (
            np.dot(self.ae.f_j, u2_j) + self.damp)
        v_g[0] = v_g[1]
        extra_data['all_electron_response'] = v_g

        # Calculate Hardness of spherical atom, for debugging purposes
        l = [np.where(f < 1e-3, e, 1000)
             for f, e in zip(self.ae.f_j, self.ae.e_j)]
        h = [np.where(f > 1e-3, e, -1000)
             for f, e in zip(self.ae.f_j, self.ae.e_j)]
        lumo_e = min(l)
        homo_e = max(h)
        if lumo_e < 999:  # If there is unoccpied orbital
            w_j = self.coefficients.get_coefficients_1d_for_lumo_perturbation()
            v_g = self.weight * np.dot(w_j, u2_j) / (
                np.dot(self.ae.f_j, u2_j) + self.damp)
            e2 = [e + np.dot(u2 * v_g, self.ae.dr)
                  for u2, e in zip(u2_j, self.ae.e_j)]
            lumo_2 = min([np.where(f < 1e-3, e, 1000)
                          for f, e in zip(self.ae.f_j, e2)])
            # print('New lumo eigenvalue:', lumo_2 * 27.2107)
            self.hardness = lumo_2 - homo_e
            # print('Hardness predicted: %10.3f eV' %
            #       (self.hardness * 27.2107))

    def write(self, writer):
        """Writes response specific data."""
        wfs = self.wfs
        kpt_comm = wfs.kd.comm
        gd = wfs.gd

        # Write the pseudodensity on the coarse grid:
        shape = (wfs.nspins,) + tuple(gd.get_size_of_global_array())
        writer.add_array('gllb_pseudo_response_potential', shape)
        if kpt_comm.rank == 0:
            for vt_G in self.vt_sG:
                writer.fill(gd.collect(vt_G) * Ha)

        def pack(X0_asp):
            X_asp = wfs.setups.empty_atomic_matrix(
                wfs.nspins, wfs.atom_partition)
            # XXX some of the provided X0_asp contain strangely duplicated
            # elements.  Take only the minimal set:
            for a in X_asp:
                X_asp[a][:] = X0_asp[a]
            return pack_atomic_matrices(X_asp)

        writer.write(gllb_atomic_density_matrices=pack(self.D_asp))
        writer.write(gllb_atomic_response_matrices=pack(self.Dresp_asp))

        writer.write(eref_s=self.eref_s)
        writer.write(eref_source_s=self.eref_source_s)

    def empty_atomic_matrix(self):
        assert self.wfs.atom_partition is self.density.atom_partition
        return self.wfs.setups.empty_atomic_matrix(self.wfs.nspins,
                                                   self.wfs.atom_partition)

    def read(self, reader):
        r = reader.hamiltonian.xc
        wfs = self.wfs

        d('Reading vt_sG')
        self.gd.distribute(r.gllb_pseudo_response_potential / reader.ha,
                           self.vt_sG)
        self.density.distribute_and_interpolate(self.vt_sG, self.vt_sg)

        def unpack(D_sP):
            return unpack_atomic_matrices(D_sP, wfs.setups)

        # Read atomic density matrices and non-local part of hamiltonian:
        D_asp = unpack(r.gllb_atomic_density_matrices)
        Dresp_asp = unpack(r.gllb_atomic_response_matrices)

        # All density matrices are loaded to all cores
        # First distribute them to match density.D_asp
        self.D_asp = self.empty_atomic_matrix()
        self.Dresp_asp = self.empty_atomic_matrix()
        for a in self.density.D_asp:
            self.D_asp[a][:] = D_asp[a]
            self.Dresp_asp[a][:] = Dresp_asp[a]
        assert len(self.Dresp_asp) == len(self.density.D_asp)

        # Further distributes the density matricies for xc-corrections
        self.Ddist_asp = self.distribute_D_asp(self.D_asp)
        self.Drespdist_asp = self.distribute_D_asp(self.Dresp_asp)

        if 'eref_s' in r:
            self.eref_s = r.eref_s
            self.eref_source_s = r.eref_source_s

    def heeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeelp(self, olddens):
        # XXX This function should be removed once the deprecated
        # `fixdensity=True` option is removed.
        from gpaw.density import redistribute_array
        self.vt_sg = redistribute_array(self.vt_sg,
                                        olddens.finegd, self.finegd,
                                        self.wfs.nspins, self.wfs.kptband_comm)
        self.vt_sG = redistribute_array(self.vt_sG,
                                        olddens.gd, self.gd,
                                        self.wfs.nspins, self.wfs.kptband_comm)



import numpy as np
from gpaw.arraydict import ArrayDict


def to_parent_comm(partition):
    # XXX assume communicator is strided, i.e. regular.
    # This actually imposes implicit limitations on things, but is not
    # "likely" to cause trouble with the usual communicators, i.e.
    # for gd/kd/bd.
    parent = partition.comm.parent
    if parent is None:
        # This should not ordinarily be necessary, but when running with
        # AtomPAW, it is.  So let's stay out of trouble.
        return partition

    members = partition.comm.get_members()
    parent_rank_a = members[partition.rank_a]

    # XXXX we hope and pray that our communicator is "equivalent" to
    # that which includes parent's rank0.
    assert min(members) == members[0]
    parent_rank_a -= members[0]  # yuckkk
    return AtomPartition(parent, parent_rank_a,
                         name='parent-%s' % partition.name)


class AtomicMatrixDistributor:
    """Class to distribute atomic dictionaries like dH_asp and D_asp."""
    def __init__(self, atom_partition, broadcast_comm,
                 work_partition=None):
        # Assumptions on communicators are as follows.
        #
        # atom_partition represents standard domain decomposition, and
        # broadcast_comm are the corresponding kpt/band communicators
        # together encompassing wfs.world.
        #
        # Initially, dH_asp are distributed over domains according to the
        # physical location of each atom, but duplicated across band
        # and k-point communicators.
        #
        # The idea is to transfer dH_asp so they are distributed equally
        # among all ranks on wfs.world, and back, when necessary.
        self.broadcast_comm = broadcast_comm

        self.grid_partition = atom_partition
        self.grid_unique_partition = to_parent_comm(self.grid_partition)

        # This represents a full distribution across grid, kpt, and band.
        if work_partition is None:
            work_partition = self.grid_unique_partition.as_even_partition()
        self.work_partition = work_partition

    def distribute(self, D_asp):
        # Right now the D are duplicated across the band/kpt comms.
        # Here we pick out a set of unique D.  With duplicates out,
        # we can redistribute one-to-one to the larger work_partition.
        # assert D_asp.partition == self.grid_partition

        Ddist_asp = ArrayDict(self.grid_unique_partition, D_asp.shapes_a,
                              dtype=D_asp.dtype)

        if self.broadcast_comm.rank != 0:
            assert len(Ddist_asp) == 0
        for a in Ddist_asp:
            Ddist_asp[a] = D_asp[a]
        Ddist_asp.redistribute(self.work_partition)
        return Ddist_asp

    def collect(self, dHdist_asp):
        # We have an array on work_partition.  We want first to
        # collect it on grid_unique_partition, the broadcast it
        # to grid_partition.

        # First receive one-to-one from everywhere.
        # assert dHdist_asp.partition == self.work_partition
        dHdist_asp = dHdist_asp.copy()
        dHdist_asp.redistribute(self.grid_unique_partition)

        dH_asp = ArrayDict(self.grid_partition, dHdist_asp.shapes_a,
                           dtype=dHdist_asp.dtype)
        if self.broadcast_comm.rank == 0:
            buf = dHdist_asp.toarray()
            assert not np.isnan(buf).any()
        else:
            buf = dH_asp.toarray()
            buf[:] = np.nan  # Let's be careful for now like --debug mode
        self.broadcast_comm.broadcast(buf, 0)
        assert not np.isnan(buf).any()
        dH_asp.fromarray(buf)
        return dH_asp


class EvenPartitioning:
    """Represents an even partitioning of N elements over a communicator.

    For example N=17 and comm.size=5 will result in this distribution:

     * rank 0 has 3 local elements: 0, 1, 2
     * rank 1 has 3 local elements: 3, 4, 5
     * rank 2 has 3 local elements: 6, 7, 8
     * rank 3 has 4 local elements: 9, 10, 11, 12
     * rank 4 has 4 local elements: 13, 14, 15, 16

    This class uses only the 'rank' and 'size' communicator attributes."""
    def __init__(self, comm, N):
        # Conventions:
        #  n, N: local/global size
        #  i, I: local/global index
        self.comm = comm
        self.N = N
        self.nlong = -(-N // comm.size)  # size of a long slice
        self.nshort = N // comm.size  # size of a short slice
        self.longcount = N % comm.size  # number of ranks with a long slice
        self.shortcount = comm.size - self.longcount  # ranks with short slice

    def nlocal(self, rank=None):
        """Get the number of locally stored elements."""
        if rank is None:
            rank = self.comm.rank
        if rank < self.shortcount:
            return self.nshort
        else:
            return self.nlong

    def minmax(self, rank=None):
        """Get the minimum and maximum index of elements stored locally."""
        if rank is None:
            rank = self.comm.rank
        I1 = self.nshort * rank
        if rank < self.shortcount:
            I2 = I1 + self.nshort
        else:
            I1 += rank - self.shortcount
            I2 = I1 + self.nlong
        return I1, I2

    def slice(self, rank=None):
        """Get the list of indices of locally stored elements."""
        I1, I2 = self.minmax(rank=rank)
        return np.arange(I1, I2)

    def global2local(self, I):
        """Get a tuple (rank, local index) from global index I."""
        nIshort = self.nshort * self.shortcount
        if I < nIshort:
            return I // self.nshort, I % self.nshort
        else:
            Ioffset = I - nIshort
            return (self.shortcount + Ioffset // self.nlong,
                    Ioffset % self.nlong)

    def local2global(self, i, rank=None):
        """Get global index I corresponding to local index i on rank."""
        if rank is None:
            rank = self.comm.rank
        return rank * self.nshort + max(rank - self.shortcount, 0) + i

    def as_atom_partition(self, strided=False, name='unnamed-even'):
        rank_a = [self.global2local(i)[0] for i in range(self.N)]
        if strided:
            rank_a = np.arange(self.comm.size).repeat(self.nlong)
            rank_a = rank_a.reshape(self.comm.size, -1).T.ravel()
            rank_a = rank_a[self.shortcount:].copy()
        return AtomPartition(self.comm, rank_a, name=name)

    def get_description(self):
        lines = []
        for a in range(self.comm.size):
            elements = ', '.join(map(str, self.slice(a)))
            line = 'rank %d has %d local elements: %s' % (a, self.nlocal(a),
                                                          elements)
            lines.append(line)
        return '\n'.join(lines)


# Interface for things that can be redistributed with general_redistribute
class Redistributable:
    def get_recvbuffer(self, a):
        raise NotImplementedError

    def get_sendbuffer(self, a):
        raise NotImplementedError

    def assign(self, a):
        raise NotImplementedError


# Let's keep this as an independent function for now in case we change the
# classes 5 times, like we do
def general_redistribute(comm, src_rank_a, dst_rank_a, redistributable):
    # To do: it should be possible to specify duplication to several ranks
    # But how is this done best?
    requests = []
    flags = (src_rank_a != dst_rank_a)
    my_incoming_atom_indices = np.argwhere(
        np.bitwise_and(flags, dst_rank_a == comm.rank)).ravel()
    my_outgoing_atom_indices = np.argwhere(
        np.bitwise_and(flags, src_rank_a == comm.rank)).ravel()

    for a in my_incoming_atom_indices:
        # Get matrix from old domain:
        buf = redistributable.get_recvbuffer(a)
        requests.append(comm.receive(buf, src_rank_a[a], tag=a, block=False))
        # These arrays are not supposed to pointers into a larger,
        # contiguous buffer, so we should make a copy - except we
        # must wait until we have completed the send/receiving
        # into them, so we will do it a few lines down.
        redistributable.assign(a, buf)

    for a in my_outgoing_atom_indices:
        # Send matrix to new domain:
        buf = redistributable.get_sendbuffer(a)
        requests.append(comm.send(buf, dst_rank_a[a], tag=a, block=False))

    comm.waitall(requests)


class AtomPartition:
    """Represents atoms distributed on a standard grid descriptor."""
    def __init__(self, comm, rank_a, name='unnamed'):
        self.comm = comm
        self.rank_a = np.array(rank_a)
        self.my_indices = self.get_indices(comm.rank)
        self.natoms = len(rank_a)
        self.name = name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AtomPartition):
            return NotImplemented
        return (self.comm.compare(other.comm) in ['ident', 'congruent']
                and np.array_equal(self.rank_a, other.rank_a))

    def __ne__(self, other):
        return not self == other

    def as_serial(self):
        return AtomPartition(self.comm, np.zeros(self.natoms, int),
                             name='%s-serial' % self.name)

    def get_indices(self, rank):
        return np.where(self.rank_a == rank)[0]

    def as_even_partition(self):
        even_part = EvenPartitioning(self.comm, len(self.rank_a))
        return even_part.as_atom_partition()

    def redistribute(self, new_partition, atomdict_ax, get_empty):
        # XXX we the two communicators to be equal according to
        # some proper criterion like MPI_Comm_compare -> MPI_IDENT.
        # But that is not implemented, so we don't.
        if self.comm.compare(new_partition.comm) not in ['ident',
                                                         'congruent']:
            msg = ('Incompatible partitions %s --> %s.  '
                   'Communicators must be at least congruent'
                   % (self, new_partition))
            raise ValueError(msg)

        # atomdict_ax may be a dictionary or a list of dictionaries

        has_many = not hasattr(atomdict_ax, 'items')
        if has_many:
            class Redist:
                def get_recvbuffer(self, a):
                    return get_empty(a)

                def assign(self, a, b_x):
                    for u, d_ax in enumerate(atomdict_ax):
                        assert a not in d_ax
                        atomdict_ax[u][a] = b_x[u]

                def get_sendbuffer(self, a):
                    return np.array([d_ax.data.pop(a) for d_ax in atomdict_ax])
        else:
            class Redist:
                def get_recvbuffer(self, a):
                    return get_empty(a)

                def assign(self, a, b_x):
                    assert a not in atomdict_ax
                    atomdict_ax[a] = b_x

                def get_sendbuffer(self, a):
                    return atomdict_ax.data.pop(a)

        try:
            general_redistribute(self.comm, self.rank_a,
                                 new_partition.rank_a, Redist())
        except ValueError as err:
            raise ValueError('redistribute %s --> %s: %s'
                             % (self, new_partition, err))
        if isinstance(atomdict_ax, ArrayDict):
            atomdict_ax.partition = new_partition  # XXX
            atomdict_ax.check_consistency()

    def __repr__(self):
        indextext = ', '.join(map(str, self.my_indices))
        return ('%s %s@rank%d/%d (%d/%d): [%s]'
                % (self.__class__.__name__, self.name, self.comm.rank,
                   self.comm.size, len(self.my_indices), self.natoms,
                   indextext))

    def arraydict(self, shapes, dtype=float):
        return ArrayDict(self, shapes, dtype)



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

from math import pi, sqrt

import numpy as np

from gpaw.gaunt import gaunt
# load points and weights for the angular integration
from gpaw.sphere.lebedev import R_nv, Y_nL, weight_n
from gpaw.spherical_harmonics import nablarlYL

r"""
                           3
             __   dn       __   __    dY
   __  2    \       L  2  \    \        L  2
  (\/n) = (  ) Y  --- )  + ) (  )  n  --- )
            /__ L dr      /__  /__  L dr
                                        v
             L            v=1    L


        dY
          L
  A   = --- r
   Lv   dr
          v

"""
# A_nvL is defined as above, n is an expansion point index (50 Lebedev points).
rnablaY_nLv = np.empty((len(R_nv), 25, 3))
for rnablaY_Lv, Y_L, R_v in zip(rnablaY_nLv, Y_nL, R_nv):
    for l in range(5):
        for L in range(l**2, (l + 1)**2):
            rnablaY_Lv[L] = nablarlYL(L, R_v) - l * R_v * Y_L[L]


class PAWXCCorrection:
    def __init__(self,
                 w_jg,    # all-lectron partial waves
                 wt_jg,   # pseudo partial waves
                 nc_g,    # core density
                 nct_g,   # smooth core density
                 rgd,     # radial grid descriptor
                 jl,      # ?
                 lmax,    # maximal angular momentum to consider
                 e_xc0,    # xc energy of reference atom
                 phicorehole_g,  # ?
                 fcorehole,      # ?
                 tauc_g,  # kinetic core energy array
                 tauct_g):  # pseudo kinetic core energy array
        self.e_xc0 = e_xc0
        self.Lmax = (lmax + 1)**2
        self.rgd = rgd
        self.dv_g = rgd.dv_g
        self.Y_nL = Y_nL[:, :self.Lmax]
        self.rnablaY_nLv = rnablaY_nLv[:, :self.Lmax]
        self.ng = ng = len(nc_g)
        self.phi_jg = w_jg
        self.phit_jg = wt_jg

        self.jlL = [(j, l, l**2 + m) for j, l in jl for m in range(2 * l + 1)]
        self.ni = ni = len(self.jlL)
        self.nj = nj = len(jl)
        self.nii = nii = ni * (ni + 1) // 2
        njj = nj * (nj + 1) // 2

        self.tauc_g = tauc_g
        self.tauct_g = tauct_g
        self.tau_npg = None
        self.taut_npg = None

        G_LLL = gaunt(max(l for j, l in jl))[:, :, :self.Lmax]
        LGcut = G_LLL.shape[2]
        B_Lqp = np.zeros((self.Lmax, njj, nii))
        p = 0
        i1 = 0
        for j1, l1, L1 in self.jlL:
            for j2, l2, L2 in self.jlL[i1:]:
                if j1 < j2:
                    q = j2 + j1 * nj - j1 * (j1 + 1) // 2
                else:
                    q = j1 + j2 * nj - j2 * (j2 + 1) // 2
                B_Lqp[:LGcut, q, p] = G_LLL[L1, L2]
                p += 1
            i1 += 1
        self.B_pqL = B_Lqp.T.copy()

        #
        self.n_qg = np.zeros((njj, ng))
        self.nt_qg = np.zeros((njj, ng))
        q = 0
        for j1, l1 in jl:
            for j2, l2 in jl[j1:]:
                self.n_qg[q] = w_jg[j1] * w_jg[j2]
                self.nt_qg[q] = wt_jg[j1] * wt_jg[j2]
                q += 1

        self.nc_g = nc_g
        self.nct_g = nct_g

        if fcorehole != 0.0:
            self.nc_corehole_g = fcorehole * phicorehole_g**2 / (4 * pi)
        else:
            self.nc_corehole_g = None

    def four_phi_integrals(self, D_sp, fxc):
        """Calculate four-phi integrals.

        The density is given by the density matrix ``D_sp`` in packed(pack)
        form, and the resulting rank-four tensor is also returned in
        packed format. ``fxc`` is a radial object???
        """

        ns, nii = D_sp.shape

        assert ns == 1

        D_p = D_sp[0]
        D_Lq = np.dot(self.B_pqL.T, D_p)

        # Expand all-electron density in spherical harmonics:
        n_qg = self.n_qg
        n_Lg = np.dot(D_Lq, n_qg)
        n_Lg[0] += self.nc_g * sqrt(4 * pi)

        # Expand pseudo electron density in spherical harmonics:
        nt_qg = self.nt_qg
        nt_Lg = np.dot(D_Lq, nt_qg)
        nt_Lg[0] += self.nct_g * sqrt(4 * pi)

        # Allocate array for result:
        J_pp = np.zeros((nii, nii))

        # Loop over 50 points on the sphere surface:
        for w, Y_L in zip(weight_n, self.Y_nL):
            B_pq = np.dot(self.B_pqL, Y_L)

            fxcdv = fxc(np.dot(Y_L, n_Lg)) * self.dv_g
            dn2_qq = np.inner(n_qg * fxcdv, n_qg)

            fxctdv = fxc(np.dot(Y_L, nt_Lg)) * self.dv_g
            dn2_qq -= np.inner(nt_qg * fxctdv, nt_qg)

            J_pp += w * np.dot(B_pq, np.inner(dn2_qq, B_pq))

        return J_pp



import numpy as np
import pytest

from gpaw.test import equal
from gpaw.grid_descriptor import GridDescriptor
from gpaw.spline import Spline
import gpaw.mpi as mpi
from gpaw.pw.descriptor import PWDescriptor
from gpaw.pw.lfc import PWLFC


@pytest.mark.ci
def test_pw_reallfc():
    x = 2.0
    rc = 3.5
    r = np.linspace(0, rc, 100)

    n = 40
    a = 8.0
    gd = GridDescriptor((n, n, n), (a, a, a), comm=mpi.serial_comm)

    spos_ac = np.array([(0.15, 0.45, 0.95)])

    pd = PWDescriptor(45, gd, complex)
    pdr = PWDescriptor(45, gd)

    for l in range(4):
        print(l)
        s = Spline(l, rc, 2 * x**1.5 / np.pi * np.exp(-x * r**2))

        lfc = PWLFC([[s]], pd)
        lfcr = PWLFC([[s]], pdr)

        c_axi = {0: np.zeros((1, 2 * l + 1), complex)}
        c_axi[0][0, 0] = 1.9
        cr_axi = {0: np.zeros((1, 2 * l + 1))}
        cr_axi[0][0, 0] = 1.9

        b = pd.zeros(1, dtype=complex)
        br = pdr.zeros(1)

        lfc.set_positions(spos_ac)
        lfc.add(b, c_axi)
        lfcr.set_positions(spos_ac)
        lfcr.add(br, cr_axi)

        a = pd.ifft(b[0])
        ar = pdr.ifft(br[0])
        equal(abs(a - ar).max(), 0, 1e-14)

        if l == 0:
            a = a[:, ::-1].copy()
            b0 = pd.fft(a)
            br0 = pdr.fft(a.real)

        lfc.integrate(b0, c_axi)
        lfcr.integrate(br0, cr_axi)

        assert abs(c_axi[0][0] - cr_axi[0][0]).max() < 1e-14

        c_axiv = {0: np.zeros((1, 2 * l + 1, 3), complex)}
        cr_axiv = {0: np.zeros((1, 2 * l + 1, 3))}
        lfc.derivative(b0, c_axiv)
        lfcr.derivative(br0, cr_axiv)
        assert abs(c_axiv[0][0] - cr_axiv[0][0]).max() < 1e-14



import pytest
from ase import Atoms
from gpaw.mpi import world
from gpaw import GPAW, FermiDirac, Mixer, MixerSum, Davidson
from gpaw.atom.all_electron import AllElectron
from gpaw.atom.generator import Generator
from gpaw.atom.configurations import parameters


@pytest.mark.slow
def test_xc_lb94(in_tmp_dir, add_cwd_to_setup_paths):
    ref1 = 'R. v. Leeuwen PhysRevA 49, 2421 (1994)'
    ref2 = 'Gritsenko IntJQuanChem 76, 407 (2000)'
    # HOMO energy in mHa for closed shell atoms
    e_HOMO_cs = {'He': 851, 'Be': 321, 'Ne': 788,
                 'Ar': 577, 'Kr': 529, 'Xe': 474,
                 'Mg': 281 + 8}

    txt = None

    print('--- Comparing LB94 with', ref1)
    print('and', ref2)

    print('**** all electron calculations')
    print('atom [refs] -e_homo diff   all in mHa')
    if world.rank == 0:
        for atom in e_HOMO_cs.keys():
            ae = AllElectron(atom, 'LB94', txt=txt)
            ae.run()
            e_homo = int(ae.e_j[-1] * 10000 + .5) / 10.
            diff = e_HOMO_cs[atom] + e_homo
            print('%2s %8g %6.1f %4.1g' %
                  (atom, e_HOMO_cs[atom], -e_homo, diff))
            assert abs(diff) < 6
    world.barrier()

    setups = {}

    print('**** 3D calculations')
    print('atom [refs] -e_homo diff   all in mHa')

    for atom in sorted(e_HOMO_cs):
        e_ref = e_HOMO_cs[atom]
        # generate setup for the atom
        if world.rank == 0 and atom not in setups:
            g = Generator(atom, 'LB94', nofiles=True, txt=txt)
            g.run(**parameters[atom])
            setups[atom] = 1
        world.barrier()

        SS = Atoms(atom, cell=(7, 7, 7), pbc=False)
        SS.center()
        c = GPAW(h=.3, xc='LB94',
                 eigensolver=Davidson(3),
                 mixer=Mixer(0.5, 7, 50.0), nbands=-2, txt=txt)
        c.calculate(SS)
        # find HOMO energy
        eps_n = c.get_eigenvalues(kpt=0, spin=0) / 27.211
        f_n = c.get_occupation_numbers(kpt=0, spin=0)
        for e, f in zip(eps_n, f_n):
            if f < 0.99:
                break
            e_homo = e
        e_homo = int(e_homo * 10000 + .5) / 10.
        diff = e_ref + e_homo
        print('%2s %8g %6.1f %4.1f' % (atom, e_ref, -e_homo, diff))
        assert abs(diff) < 7

    # HOMO energy in mHa and magn. mom. for open shell atoms
    e_HOMO_os = {'He': [851, 0],  # added for cross check
                 'H': [440, 1],
                 'N': [534 - 23, 3],
                 'Na': [189 + 17, 1],
                 'P': [385 - 16, 3]}

    for atom in sorted(e_HOMO_os):
        e_ref = e_HOMO_os[atom][0]
        magmom = e_HOMO_os[atom][1]
        # generate setup for the atom
        if world.rank == 0 and atom not in setups:
            g = Generator(atom, 'LB94', nofiles=True, txt=txt)
            g.run(**parameters[atom])
            setups[atom] = 1
        world.barrier()

        SS = Atoms(atom, magmoms=[magmom], cell=(7, 7, 7), pbc=False)
        SS.center()
        # fine grid needed for convergence!
        c = GPAW(h=0.2,
                 xc='LB94',
                 nbands=-2,
                 spinpol=True,
                 hund=True,
                 eigensolver=Davidson(3),
                 mixer=MixerSum(0.5, 7, 50.0),
                 occupations=FermiDirac(0.0, fixmagmom=True),
                 txt=txt)
        c.calculate(SS)
        # find HOMO energy
        eps_n = c.get_eigenvalues(kpt=0, spin=0) / 27.211
        f_n = c.get_occupation_numbers(kpt=0, spin=0)
        for e, f in zip(eps_n, f_n):
            if f < 0.99:
                break
            e_homo = e
        e_homo = int(e_homo * 10000 + .5) / 10.
        diff = e_ref + e_homo
        print('%2s %8g %6.1f %4.1f' % (atom, e_ref, -e_homo, diff))
        assert abs(diff) < 15



import sys

from ase import Atoms
from ase.utils.timing import Timer
import pytest

from gpaw import GPAW
from gpaw.test import equal
from gpaw.xc.hybrid import HybridXC


@pytest.mark.libxc
@pytest.mark.hybrids
def test_exx_coarse(in_tmp_dir):
    timer = Timer()

    loa = Atoms('Be2',
                [(0, 0, 0), (2.45, 0, 0)],
                magmoms=[0.5, 0.5],
                cell=[5.9, 4.8, 5.0])
    loa.center()

    fgl = [False, True]
    # fgl = [True, False]

    E = {}
    niter = {}
    for fg in fgl:
        if fg:
            tstr = 'Exx on fine grid'
        else:
            tstr = 'Exx on coarse grid'
        timer.start(tstr)
        loa.calc = GPAW(h=0.3,
                        eigensolver='rmm-diis',
                        xc=dict(name='PBE', stencil=1),
                        poissonsolver={'name': 'fd'},
                        nbands=4,
                        convergence={'eigenstates': 1e-4},
                        charge=-1)
        E[fg] = loa.get_potential_energy()
        loa.calc = loa.calc.new(xc=HybridXC('PBE0', stencil=1, finegrid=fg))
        E[fg] = loa.get_potential_energy()
        niter[fg] = loa.calc.get_number_of_iterations()
        timer.stop(tstr)
        if not fg:
            fname = 'exx_load.gpw'
            loa.calc.write(fname)
            calcl = GPAW(fname)
            func = calcl.parameters.xc

            assert func['name'] == 'PBE0', 'wrong name for functional'
            assert func['hybrid'] == 0.25, 'wrong factor for functional'

    timer.write(sys.stdout)

    print('Total energy on the fine grid   =', E[True])
    print('Total energy on the coarse grid =', E[False])
    equal(E[True], E[False], 0.01)

    energy_tolerance = 0.003
    equal(E[False], 6.97818, energy_tolerance)
    equal(E[True], 6.97153, energy_tolerance)



from __future__ import annotations

import numpy as np
from gpaw.core.atom_arrays import AtomArrays
from gpaw.core.matrix import Matrix, create_distribution
from gpaw.core.plane_waves import (PlaneWaveAtomCenteredFunctions,
                                   PlaneWaveExpansions, PlaneWaves)
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions
from gpaw.typing import Array2D
from gpaw.new.ibzwfs import IBZWaveFunctions
from gpaw.new.wave_functions import WaveFunctions
from gpaw.new.potential import Potential
from gpaw.new.smearing import OccupationNumberCalculator


def pw_matrix(pw: PlaneWaves,
              pt_aiG: PlaneWaveAtomCenteredFunctions,
              dH_aii: AtomArrays,
              dS_aii: list[Array2D],
              vt_R: UniformGridFunctions,
              comm) -> tuple[Matrix, Matrix]:
    """Calculate H and S matrices in plane-wave basis.

    :::

                 _ _     _ _
            /  -iG.r ~  iG.r _
      O   = | e      O e    dr
       GG'  /

    :::

      ~   ^   ~ _    _ _     --- ~a _ _a    a  ~  _  _a
      H = T + v(r) δ(r-r') + <   p (r-R ) ΔH   p (r'-R )
                             ---  i         ij  j
                             aij

    :::

      ~     _ _     --- ~a _ _a    a  ~  _  _a
      S = δ(r-r') + <   p (r-R ) ΔS   p (r'-R )
                    ---  i         ij  j
                    aij
    """
    assert pw.dtype == complex
    npw = pw.shape[0]
    dist = create_distribution(npw, npw, comm, -1, 1)
    H_GG = dist.matrix(complex)
    S_GG = dist.matrix(complex)
    G1, G2 = dist.my_row_range()

    x_G = pw.empty()
    assert isinstance(x_G, PlaneWaveExpansions)  # Fix this!
    x_R = vt_R.desc.new(dtype=complex).zeros()
    assert isinstance(x_R, UniformGridFunctions)  # Fix this!
    dv = pw.dv

    for G in range(G1, G2):
        x_G.data[:] = 0.0
        x_G.data[G] = 1.0
        x_G.ifft(out=x_R)
        x_R.data *= vt_R.data
        x_R.fft(out=x_G)
        H_GG.data[G - G1] = dv * x_G.data

    H_GG.add_to_diagonal(dv * pw.ekin_G)
    S_GG.data[:] = 0.0
    S_GG.add_to_diagonal(dv)

    pt_aiG._lazy_init()
    assert pt_aiG._lfc is not None
    f_GI = pt_aiG._lfc.expand()
    nI = f_GI.shape[1]
    dH_II = np.zeros((nI, nI))
    dS_II = np.zeros((nI, nI))
    I1 = 0
    for a, dH_ii in dH_aii.items():
        dS_ii = dS_aii[a]
        I2 = I1 + len(dS_ii)
        dH_II[I1:I2, I1:I2] = dH_ii
        dS_II[I1:I2, I1:I2] = dS_ii
        I1 = I2

    H_GG.data += (f_GI[G1:G2].conj() @ dH_II) @ f_GI.T
    S_GG.data += (f_GI[G1:G2].conj() @ dS_II) @ f_GI.T

    return H_GG, S_GG


def diagonalize(potential: Potential,
                ibzwfs: IBZWaveFunctions,
                occ_calc: OccupationNumberCalculator,
                nbands: int | None) -> IBZWaveFunctions:
    """Diagonalize hamiltonian in plane-wave basis."""
    vt_sR = potential.vt_sR
    dH_asii = potential.dH_asii

    if nbands is None:
        nbands = min(wfs.array_shape(global_shape=True)[0]
                     for wfs in ibzwfs)
        array = np.array(nbands)
        ibzwfs.kpt_comm.max(array)
        nbands = int(array)

    wfs_qs: list[list[WaveFunctions]] = []
    for wfs_s in ibzwfs.wfs_qs:
        wfs_qs.append([])
        for wfs in wfs_s:
            dS_aii = [setup.dO_ii for setup in wfs.setups]
            assert isinstance(wfs, PWFDWaveFunctions)
            assert isinstance(wfs.pt_aiX, PlaneWaveAtomCenteredFunctions)
            H_GG, S_GG = pw_matrix(wfs.psit_nX.desc,
                                   wfs.pt_aiX,
                                   dH_asii[:, wfs.spin],
                                   dS_aii,
                                   vt_sR[wfs.spin],
                                   wfs.psit_nX.comm)
            eig_n = H_GG.eigh(S_GG, limit=nbands)
            if eig_n[0] < -1000:
                raise RuntimeError(
                    f'Lowest eigenvalue is {eig_n[0]} Hartree. '
                    'You might be suffering from MKL library bug MKLD-11440. '
                    'See issue #241 in GPAW. '
                    'Creashing to prevent corrupted results.')
            psit_nG = wfs.psit_nX.desc.empty(nbands)
            psit_nG.data[:nbands] = H_GG.data[:nbands].conj()
            new_wfs = PWFDWaveFunctions(
                psit_nG,
                wfs.spin,
                wfs.q,
                wfs.k,
                wfs.setups,
                wfs.fracpos_ac,
                wfs.atomdist,
                wfs.weight,
                wfs.ncomponents)
            new_wfs._eig_n = eig_n
            wfs_qs[-1].append(new_wfs)

    new_ibzwfs = IBZWaveFunctions(
        ibzwfs.ibz,
        ibzwfs.nelectrons,
        ibzwfs.ncomponents,
        wfs_qs,
        ibzwfs.kpt_comm)

    new_ibzwfs.calculate_occs(occ_calc)

    return new_ibzwfs



import numpy as np


def cholesky(a):
    from gpaw.gpu import cupy as cp
    return cp.ndarray(np.linalg.cholesky(a._data))


def inv(a):
    from gpaw.gpu import cupy as cp
    return cp.ndarray(np.linalg.inv(a._data))


def eigh(a, UPLO):
    from gpaw.gpu import cupy as cp
    eigvals, eigvecs = np.linalg.eigh(a._data, UPLO)
    return cp.ndarray(eigvals), cp.ndarray(eigvecs.T.copy().T)



"""Post processing module for dfpt force constants."""

import os
import pickle

import numpy as np
import numpy.fft as fft

import ase.phonons as phonons

from gpaw.kpt_descriptor import KPointDescriptor
from gpaw import debug


class Phonons(phonons.Phonons):
    """DFPT version of the ``Phonons`` class from ase.

    This class recycle the functionality from the ASE class.

    """

    def __init__(self, atoms, kpts, symmetry):
        """Initialize base class and attributes.

        Parameters
        ----------
        atoms: Atoms
            ASE atoms.
        kpts: tuple or list of tuples
            Shape of Monkhorst-Pack grid or list of k-points used in the dfpt
            calculation.
        symmetry: bool or None
            Symmetry parameter used in dfpt calculation.

        """

        # Init base class with ``Atoms`` object
        phonons.Phonons.__init__(atoms)

        # Create k-point descriptor
        self.kd = KPointDescriptor(kpts, 1)
        self.kd.set_symmetry(self.atoms, symmetry)

        # Overwrite ``N_c`` attribute
        self.N_c = tuple(self.kd.N_c)

        # Index of the gamma point -- for the acoustic sum-rule
        self.gamma_index = None
        if self.kd.gamma:
            self.gamma_index = 0
            self.dtype == float
        else:
            self.dtype == complex
            for k, k_c in enumerate(self.kd.ibzk_kc):
                if np.all(k_c == 0.):
                    self.gamma_index = k

        assert self.gamma_index is not None

    def run(self):
        """Overwrite base class member function."""

        raise RuntimeError("Use only this class for post-processing")

    def read(self):
        """Read force constants from files."""

        # Data structure for force constants
        self.C_qaavv = [dict([(a, dict([(a_, np.zeros((3, 3),
                                                      dtype=self.dtype))
                                        for a_ in self.indices]))
                              for a in self.indices])
                        for q in range(self.kd.nibzkpts)]

        assert self.name is not None
        assert self.path is not None

        for q in range(self.kd.nibzkpts):
            for a in self.indices:
                for v in [0, 1, 2]:
                    filename = self.name % (q, a, v)
                    try:
                        fd = open(os.path.join(self.path, filename), 'rb')
                    except EOFError:
                        print(("Redo file %s "
                               % os.path.join(self.path, filename)))
                    C_qav_a = pickle.load(fd)
                    fd.close()
                    for a_ in self.indices:
                        self.C_qaavv[q][a][a_][v] = C_qav_a[a_]

    def assemble(self, acoustic=True):
        """Assemble dynamical matrix from the force constants attribute.

        The elements of the dynamical matrix are given by::

            D_ij(q) = 1/(M_i + M_j) * C_ij(q) ,

        where i and j are collective atomic and cartesian indices.

        During the assembly, various symmetries of the dynamical matrix are
        enforced::

            1) Hermiticity
            2) Acoustic sum-rule
            3) D(q) = D*(-q)

        Parameters
        ----------
        acoustic: bool
            When True, the diagonal of the matrix of force constants is
            corrected to ensure that the acoustic sum-rule is fulfilled.

        """

        # Read force constants from files
        self.read()

        # Number of atoms included
        N = len(self.indices)

        # Assemble matrix of force constants
        self.C_q = []
        for q, C_aavv in enumerate(self.C_qaavv):

            C_avav = np.zeros((3 * N, 3 * N), dtype=self.dtype)

            for i, a in enumerate(self.indices):
                for j, a_ in enumerate(self.indices):
                    C_avav[3 * i: 3 * i + 3, 3 * j: 3 * j + 3] += C_aavv[a][a]

            self.C_q.append(C_avav)

        # XXX Figure out in which order the corrections should be done
        # Make C(q) Hermitian
        for C in self.C_q:
            C *= 0.5
            C += C.T.conj()

        # Get matrix of force constants in the Gamma-point (is real!)
        C_gamma = self.C_q[self.gamma_index].real
        # Make Gamma-component real
        self.C_q[self.gamma_index] = C_gamma.copy()

        # Apply acoustic sum-rule if requested
        if acoustic:
            # Correct atomic diagonal for each q-vector
            for C in self.C_q:
                for a in range(N):
                    for a_ in range(N):
                        C[3 * a: 3 * a + 3, 3 * a: 3 * a + 3] -= \
                            C_gamma[3 * a: 3 * a + 3, 3 * a_: 3 * a_ + 3]

            # Check sum-rule for Gamma-component in debug mode
            if debug:
                C = self.C_q[self.gamma_index]
                assert np.all(np.sum(C.reshape((3 * N, N, 3)), axis=1) < 1e-15)

        # Move this bit to an ``unfold`` member function
        # XXX Time-reversal symmetry
        C_q = np.asarray(self.C_q)
        if self.kd.nibzkpts != self.kd.nbzkpts:
            self.D_k = np.concatenate((C_q[:0:-1].conjugate(), C_q))
        else:
            self.D_k = 0.5 * C_q
            self.D_k += self.D_k[::-1].conjugate()

        # Mass prefactor for the dynamical matrix
        m = self.atoms.get_masses()
        self.m_inv_av = np.repeat(m[self.indices]**-0.5, 3)
        M_avav = self.m_inv_av[:, np.newaxis] * self.m_inv_av

        for C in self.D_k:
            C *= M_avav

        self.assembled = True

    def real_space(self):
        """Fourier transform the dynamical matrix to real-space."""

        if not self.assembled:
            self.assemble()

        # Shape of q-point grid
        N_c = self.N_c

        # Reshape before Fourier transforming
        shape = self.D_k.shape
        Dq_lmn = self.D_k.reshape(N_c + shape[1:])
        DR_lmn = fft.ifftn(fft.ifftshift(Dq_lmn, axes=(0, 1, 2)),
                           axes=(0, 1, 2))

        if debug:
            # Check that D_R is real enough
            assert np.all(DR_lmn.imag < 1e-8)

        DR_lmn = DR_lmn.real

        # Corresponding R_m vectors in units of the basis vectors
        R_cm = np.indices(N_c).reshape(3, -1)
        N1_c = np.array(N_c)[:, np.newaxis]
        R_cm += N1_c // 2
        R_cm %= N1_c
        R_cm -= N1_c // 2
        R_clmn = R_cm.reshape((3,) + N_c)

        return DR_lmn, R_clmn

    def fourier_interpolate(self, N_c):
        """Fourier interpolate dynamical matrix onto a finer q-vector grid."""

        # Move this member function to the ASE class
        raise NotImplementedError



import numpy as np
from ase.units import Bohr, _hbar, _e, _me, _eps0

from gpaw.nlopt.basic import load_data
from gpaw.mpi import world


def get_chi_tensor(
        freqs=[1.0], eta=0.05,
        ftol=1e-4, Etol=1e-6, eshift=0.0,
        band_n=None, mml_name='mml.npz', out_name=None):
    """
    Calculate full linear susceptibility tensor for nonmagnetic semiconductors.

    Input:
        freqs           Excitation frequency array (a numpy array or list)
        eta             Broadening, a number or an array (default 0.05 eV)
        Etol, ftol      Tol. in energy and fermi to consider degeneracy
        eshift          Bandgap correction
        band_n          List of bands in the sum (default 0 to nb)
        out_name        If it is given: output filename
        mml_name        The momentum filename (default 'mml.npz')
    Output:
        chi_vvl         The output tensor (3, 3, nw)
        chi.npy         If specified: array containing the spectrum and freqs

    """

    freqs = np.array(freqs)
    nw = len(freqs)
    w_lc = freqs + 1j * eta

    # Load the required data
    k_info = load_data(mml_name=mml_name)
    if k_info:
        tmp = list(k_info.values())[0]
        nb = len(tmp[1])
        if band_n is None:
            band_n = list(range(nb))

    # Initialize the outputs
    sum_vvl = np.zeros((3, 3, nw), complex)

    # Do the calculations
    for _, (we, f_n, E_n, p_vnn) in k_info.items():
        tmp = np.zeros((3, 3, nw), complex)
        for v1 in range(3):
            for v2 in range(v1, 3):
                sum_l = calc_chi(
                    w_lc, f_n, E_n, p_vnn, [v1, v2],
                    band_n, ftol, Etol, eshift=eshift)
                tmp[v1, v2] = sum_l
                tmp[v2, v1] = sum_l
        # Add it to previous with a weight
        sum_vvl += tmp * we

    world.sum(sum_vvl)

    # Make the output in SI unit
    dim_sigma = 1j * _e**2 * _hbar / (_me**2 * (2 * np.pi)**3)
    dim_chi = 1j * _hbar / (_eps0 * _e)
    dim_sum = (_hbar / (Bohr * 1e-10))**2 / \
        (_e**2 * (Bohr * 1e-10)**3)
    dim_SI = dim_sigma * dim_chi * dim_sum
    chi_vvl = dim_SI * sum_vvl

    # Save it to the file
    if world.rank == 0 and out_name is not None:
        tmp = chi_vvl.reshape(9, nw)
        lin = np.vstack((freqs, tmp))
        np.save(out_name, lin)

    return chi_vvl


def calc_chi(
        w_l, f_n, E_n, p_vnn, pol_v,
        band_n=None, ftol=1e-4, Etol=1e-6, eshift=0):
    """
    Loop over bands for computing the response

    Input:
        w_l             Complex frequency array
        f_n             Fermi levels
        E_n             Energies
        p_vnn           Momentum matrix elements
        pol_v           Tensor element
        band_n          Band list
        Etol, ftol      Tol. in energy and fermi to consider degeneracy
        eshift          Bandgap correction
    Output:
        sum_l           Output sum value (array with length of w_l)
    """

    # Initialize variables
    nb = len(f_n)
    if band_n is None:
        band_n = list(range(nb))
    sum_l = np.zeros(w_l.size, complex)

    # Loop over bands
    for nni in band_n:
        for mmi in band_n:
            # Use TRS to reduce
            if mmi <= nni:
                continue
            fnm = f_n[nni] - f_n[mmi]
            Emn = E_n[mmi] - E_n[nni] + fnm * eshift
            if np.abs(fnm) < ftol or np.abs(Emn) < Etol:
                continue
            # *2 for real, /2 for TRS, *2 for m<n
            sum_l += 2 * fnm * np.real(
                p_vnn[pol_v[0], nni, mmi] * p_vnn[pol_v[1], mmi, nni]) \
                / (Emn * (w_l**2 - Emn**2))

    return sum_l



"""This module implements a linear response calculator class."""

__all__ = ["ResponseCalculator"]

import numpy as np

from gpaw.transformers import Transformer
from gpaw.dfpt.mixer import BaseMixer
from gpaw.dfpt.sternheimeroperator import SternheimerOperator
from gpaw.dfpt.scipylinearsolver import ScipyLinearSolver
from gpaw.dfpt.preconditioner import ScipyPreconditioner


class ResponseCalculator:
    """This class is a calculator for the sc density variation.

    From the given perturbation, the set of coupled equations for the
    first-order density response is solved self-consistently.

    Parameters
    ----------
    max_iter: int
        Maximum number of iterations in the self-consistent evaluation of
        the density variation.
    tolerance_sc: float
        Tolerance for the self-consistent loop measured in terms of
        integrated absolute change of the density derivative between two
        iterations.
    tolerance_sternheimer: float
        Tolerance for the solution of the Sternheimer equation -- passed to
        the ``LinearSolver``.
    beta: float (0 < beta < 1)
        Mixing coefficient.
    nmaxold: int
        Length of history for the mixer.
    weight: int
        Weight for the mixer metric (=1 -> no metric used).

    """

    parameters = {'verbose': False,
                  'max_iter': 100,
                  'max_iter_krylov': 1000,
                  'krylov_solver': 'cg',
                  'tolerance_sc': 1.0e-5,
                  'tolerance_sternheimer': 1.0e-4,
                  'use_pc': True,
                  'beta': 0.1,
                  'nmaxold': 6,
                  'weight': 50}

    def __init__(self, calc, wfs, poisson_solver=None, dtype=float, **kwargs):
        """Store calculator etc.

        Parameters
        ----------
        calc: Calculator
            Calculator instance containing a ground-state calculation
            (calc.set_positions must have been called before this point!).
        wfs: WaveFunctions
            Class taking care of wave-functions, projectors, k-point related
            quantities and symmetries.
        poisson_solver: PoissonSolver
            Multigrid or FFT poisson solver (not required if the
            ``Perturbation`` to be solved for has a ``solve_poisson`` member
            function).
        dtype: ...
            dtype of the density response.

        """

        # Store ground-state quantities
        self.hamiltonian = calc.hamiltonian
        self.density = calc.density

        self.wfs = wfs

        # Get list of k-point containers
        self.kpt_u = wfs.kpt_u

        # Poisson solver
        if poisson_solver is None:
            # Solver must be provided by the perturbation
            self.poisson = None
            self.solve_poisson = None
        else:
            self.poisson = poisson_solver
            self.solve_poisson = self.poisson.solve_neutral

        # Store grid-descriptors
        self.gd = calc.density.gd
        self.finegd = calc.density.finegd

        # dtype for ground-state wave-functions
        self.gs_dtype = calc.wfs.dtype
        # dtype for the perturbing potential and density
        self.dtype = dtype

        # Grid transformer -- convert array from coarse to fine grid
        self.interpolator = Transformer(self.gd, self.finegd, nn=3,
                                        dtype=self.dtype)
        # Grid transformer -- convert array from fine to coarse grid
        self.restrictor = Transformer(self.finegd, self.gd, nn=3,
                                      dtype=self.dtype)

        # Sternheimer operator
        self.sternheimer_operator = None
        # Krylov solver
        self.linear_solver = None

        # Phases for transformer objects - since the perturbation determines
        # the form of the density response this is obtained from the
        # perturbation in the ``__call__`` member function below.
        self.phase_cd = None

        # Array attributes
        self.nt1_G = None
        self.vHXC1_G = None
        self.nt1_g = None
        self.vH1_g = None

        # Perturbation
        self.perturbation = None

        # Number of occupied bands
        nvalence = calc.wfs.nvalence
        self.nbands = nvalence // 2 + nvalence % 2
        assert self.nbands <= calc.wfs.bd.nbands

        self.initialized = False

        self.parameters = {}
        self.set(**kwargs)

    def clean(self):
        """Cleanup before call to ``__call__``."""

        self.perturbation = None
        self.solve_poisson = None

        self.nt1_G = None
        self.vHXC1_G = None
        self.nt1_g = None
        self.vH1_g = None

    def __call__(self, perturbation):
        """Calculate density response (derivative) to perturbation.

        Parameters
        ----------
        perturbation: Perturbation
            Class implementing the perturbing potential. Must provide an
            ``apply`` member function implementing the multiplication of the
            perturbing potential to a (set of) state vector(s).

        """

        assert self.initialized, ("Linear response calculator "
                                  "not initizalized.")
        self.clean()

        if self.poisson is None:
            assert hasattr(perturbation, 'solve_poisson')
            self.solve_poisson = perturbation.solve_poisson

        # Store perturbation - used in other member functions
        self.perturbation = perturbation
        # Reset mixer
        self.mixer.reset()
        # Reset wave-functions
        self.wfs.reset()

        # Set phase attribute for Transformer objects
        self.phase_cd = self.perturbation.get_phase_cd()

        # Parameters for the SC-loop
        p = self.parameters
        max_iter = p['max_iter']
        tolerance = p['tolerance_sc']

        for iter in range(max_iter):

            if iter == 0:
                self.first_iteration()
            else:
                print("iter:%3i\t" % iter, end=' ')
                norm = self.iteration()
                print("abs-norm: %6.3e\t" % norm, end=' ')
                print(("integrated density response (abs): % 5.2e (%5.2e) "
                       % (self.gd.integrate(self.nt1_G.real),
                          self.gd.integrate(np.absolute(self.nt1_G)))))

                if norm < tolerance:
                    print(("self-consistent loop converged in %i iterations"
                           % iter))
                    break

            if iter == max_iter:
                raise RuntimeError("self-consistent loop did not converge "
                                   "in %i iterations" % iter)

    def set(self, **kwargs):
        """Set parameters for calculation."""

        # Check for legal input parameters
        for key, value in kwargs.items():
            if key not in ResponseCalculator.parameters:
                raise TypeError("Unknown keyword argument: '%s'" % key)

        # Insert default values if not given
        for key, value in ResponseCalculator.parameters.items():
            if key not in kwargs:
                kwargs[key] = value

        self.parameters.update(kwargs)

    def initialize(self, spos_ac):
        """Make the object ready for a calculation."""

        # Parameters
        p = self.parameters
        beta = p['beta']
        nmaxold = p['nmaxold']
        weight = p['weight']
        # use_pc = p['use_pc']
        tolerance_sternheimer = p['tolerance_sternheimer']
        max_iter_krylov = p['max_iter_krylov']
        krylov_solver = p['krylov_solver']

        # Initialize WaveFunctions attribute
        self.wfs.initialize(spos_ac)

        # Initialize mixer
        # weight = 1 -> no metric is used
        self.mixer = BaseMixer(beta=beta, nmaxold=nmaxold,
                               weight=weight, dtype=self.dtype)
        self.mixer.initialize_metric(self.gd)

        # Linear operator in the Sternheimer equation
        self.sternheimer_operator = \
            SternheimerOperator(self.hamiltonian, self.wfs, self.gd,
                                dtype=self.gs_dtype)

        # Preconditioner for the Sternheimer equation
        if p['use_pc']:
            pc = ScipyPreconditioner(self.gd,
                                     self.sternheimer_operator.project,
                                     dtype=self.gs_dtype)
        else:
            pc = None

        # XXX K-point of the pc must be set in the k-point loop -> store a ref.
        self.pc = pc
        # Linear solver for the solution of Sternheimer equation
        self.linear_solver = ScipyLinearSolver(method=krylov_solver,
                                               preconditioner=pc,
                                               tolerance=tolerance_sternheimer,
                                               max_iter=max_iter_krylov)

        self.initialized = True

    def first_iteration(self):
        """Perform first iteration of sc-loop."""

        self.wave_function_variations()
        self.density_response()
        self.mixer.mix(self.nt1_G, [], phase_cd=self.phase_cd)
        self.interpolate_density()

    def iteration(self):
        """Perform iteration."""

        # Update variation in the effective potential
        self.effective_potential_variation()
        # Update wave function variations
        self.wave_function_variations()
        # Update density
        self.density_response()
        # Mix - supply phase_cd here for metric inside the mixer
        norm = self.mixer.mix(self.nt1_G, [], phase_cd=self.phase_cd)

        self.interpolate_density()

        return norm

    def interpolate_density(self):
        """Interpolate density derivative onto the fine grid."""

        self.nt1_g = self.finegd.zeros(dtype=self.dtype)
        self.interpolator.apply(self.nt1_G, self.nt1_g, phases=self.phase_cd)

    def effective_potential_variation(self):
        """Calculate derivative of the effective potential (Hartree + XC)."""

        # Hartree part
        vHXC1_g = self.finegd.zeros(dtype=self.dtype)
        self.solve_poisson(vHXC1_g, self.nt1_g)
        # Store for evaluation of second order derivative
        self.vH1_g = vHXC1_g.copy()

        # XC part
        nt_sg = self.density.nt_sg
        fxct_sg = np.zeros_like(nt_sg)
        self.hamiltonian.xc.calculate_fxc(self.finegd, nt_sg, fxct_sg)
        vHXC1_g += fxct_sg[0] * self.nt1_g

        # Transfer to coarse grid
        self.vHXC1_G = self.gd.zeros(dtype=self.dtype)
        self.restrictor.apply(vHXC1_g, self.vHXC1_G, phases=self.phase_cd)

    def wave_function_variations(self):
        """Calculate variation in the wave-functions.

        Parameters
        ----------
        v1_G: ndarray
            Variation of the local effective potential (Hartree + XC).

        """

        verbose = self.parameters['verbose']

        if verbose:
            print("Calculating wave function variations")

        if self.perturbation.has_q():
            q_c = self.perturbation.get_q()
            kplusq_k = self.wfs.kd.find_k_plus_q(q_c)
        else:
            kplusq_k = None

        # Calculate wave-function variations for all k-points.
        for kpt in self.kpt_u:

            k = kpt.k

            if verbose:
                print("k-point %2.1i" % k)

            # Index of k+q vector
            if kplusq_k is None:
                kplusq = k
                kplusqpt = kpt
            else:
                kplusq = kplusq_k[k]
                kplusqpt = self.kpt_u[kplusq]

            # Ground-state and first-order wave-functions
            psit_nG = kpt.psit_nG
            psit1_nG = kpt.psit1_nG
            # Update the SternheimerOperator
            self.sternheimer_operator.set_k(k)
            self.sternheimer_operator.set_kplusq(kplusq)
            # Update preconditioner
            if self.pc is not None:
                # k+q
                self.pc.set_kpt(kplusqpt)

            # Right-hand side of Sternheimer equations
            # k and k+q
            # XXX should only be done once for all k-points but maybe too cheap
            # to bother ??
            rhs_nG = self.gd.zeros(n=self.nbands, dtype=self.gs_dtype)
            self.perturbation.apply(psit_nG, rhs_nG, self.wfs, k, kplusq)
            if self.vHXC1_G is not None:
                rhs_nG += self.vHXC1_G * psit_nG
            # Project out occupied subspace
            self.sternheimer_operator.project(rhs_nG)

            # Loop over occupied bands
            for n in range(self.nbands):

                # Update band index in SternheimerOperator
                self.sternheimer_operator.set_band(n)
                # Get view of the Bloch function derivative
                psit1_G = psit1_nG[n]
                # Rhs of Sternheimer equation
                rhs_G = -1 * rhs_nG[n]

                # Solve Sternheimer equation
                iter, info = self.linear_solver.solve(
                    self.sternheimer_operator,
                    psit1_G, rhs_G)

                if verbose:
                    print("\tBand %2.1i -" % n, end=' ')

                if info == 0:
                    if verbose:
                        print("linear solver converged in %i iterations" %
                              iter)
                elif info > 0:
                    assert False, ("linear solver did not converge in maximum "
                                   "number (=%i) of iterations for "
                                   "k-point number %d" % (iter, k))
                else:
                    assert False, ("linear solver failed to converge")

    def density_response(self):
        """Calculate density response from variation in the wave-functions."""

        # Density might be complex
        self.nt1_G = self.gd.zeros(dtype=self.dtype)

        for kpt in self.kpt_u:
            # The weight of the k-points includes spin-degeneracy
            w = kpt.weight
            # Wave functions
            psit_nG = kpt.psit_nG
            psit1_nG = kpt.psit1_nG

            for psit_G, psit1_G in zip(psit_nG, psit1_nG):
                # NOTICE: this relies on the automatic down-cast of the complex
                # array on the rhs to a real array when the lhs is real !!
                # Factor 2 for time-reversal symmetry
                self.nt1_G += 2 * w * psit_G.conj() * psit1_G



"""Part of the module for electrodynamic simulations

"""


class PotentialCoupler():
    def __init__(self,
                 qm,  # PoissonOrganizer
                 cl,  # PoissonOrganizer
                 index_offset_1,
                 index_offset_2,
                 extended_index_offset_1,
                 extended_index_offset_2,
                 extended_delta_index,
                 num_refinements,
                 remove_moment_qm,
                 remove_moment_cl,
                 rank):
        self.qm = qm
        self.cl = cl
        self.index_offset_1 = index_offset_1
        self.index_offset_2 = index_offset_2
        self.extended_index_offset_1 = extended_index_offset_1
        self.extended_index_offset_2 = extended_index_offset_2
        self.extended_delta_index = extended_delta_index
        self.num_refinements = num_refinements
        self.remove_moment_qm = remove_moment_qm
        self.remove_moment_cl = remove_moment_cl
        self.rank = rank

        # These are used to remember the previous solutions
        self.old_local_phi_qm_qmgd = self.qm.gd.zeros()
        self.old_local_phi_cl_clgd = self.cl.gd.zeros()
        self.old_local_phi_qm_clgd = self.cl.gd.zeros()

    # Add quantum and classical potentials
    def couple(self,
               qm_qmgd,
               qm_clgd,
               cl_qmgd,
               cl_clgd):
        tot_qmgd = cl_qmgd + qm_qmgd
        tot_clgd = cl_clgd + qm_clgd
        return tot_qmgd, tot_clgd


# Coarsen quantum charge to suit the classical grid,
# and refine classical potential to suit the quantum grid
class RefinerPotentialCoupler(PotentialCoupler):

    def __init__(self,
                 qm,
                 cl,
                 index_offset_1,
                 index_offset_2,
                 extended_index_offset_1,
                 extended_index_offset_2,
                 extended_delta_index,
                 num_refinements,
                 remove_moment_qm,
                 remove_moment_cl,
                 rank):
        PotentialCoupler.__init__(self,
                                  qm,
                                  cl,
                                  index_offset_1,
                                  index_offset_2,
                                  extended_index_offset_1,
                                  extended_index_offset_2,
                                  extended_delta_index,
                                  num_refinements,
                                  remove_moment_qm,
                                  remove_moment_cl,
                                  rank)

    def getPotential(self, local_rho_qm_qmgd, local_rho_cl_clgd, **kwargs):
        # Quantum potential
        local_phi_qm_qmgd = self.old_local_phi_qm_qmgd
        niter_qm = self.qm.poisson_solver.solve(phi=local_phi_qm_qmgd,
                                                rho=local_rho_qm_qmgd,
                                                **kwargs)
        self.old_local_phi_qm_qmgd = local_phi_qm_qmgd.copy()

        # Classical potential
        local_phi_cl_clgd = self.old_local_phi_cl_clgd
        niter_cl = self.cl.poisson_solver.solve(phi=local_phi_cl_clgd,
                                                rho=local_rho_cl_clgd,
                                                **kwargs)
        self.old_local_phi_cl_clgd = local_phi_cl_clgd.copy()

        # Transfer classical potential into quantum subsystem
        global_phi_cl_clgd = self.cl.gd.collect(local_phi_cl_clgd)
        local_phi_cl_qmgd = self.qm.gd.zeros()
        global_phi_cl_qmgd = self.qm.gd.zeros(global_array=True)

        if self.rank == 0:
            # Extract the overlapping region from the classical potential
            o1 = self.extended_index_offset_1
            o2 = self.extended_index_offset_2
            global_phi_cl_clgd_refined = global_phi_cl_clgd[
                o1[0]:o2[0] - 1,
                o1[1]:o2[1] - 1,
                o1[2]:o2[2] - 1].copy()

            for n in range(self.num_refinements):
                global_phi_cl_clgd_refined = \
                    self.cl.extended_refiners[n].apply(
                        global_phi_cl_clgd_refined)

            global_phi_cl_qmgd = global_phi_cl_clgd_refined[
                self.extended_delta_index:-self.extended_delta_index,
                self.extended_delta_index:-self.extended_delta_index,
                self.extended_delta_index:-self.extended_delta_index]

        self.qm.gd.distribute(global_phi_cl_qmgd, local_phi_cl_qmgd)

        # Transfer quantum density into classical grid
        global_rho_qm_qmgd = self.qm.gd.collect(local_rho_qm_qmgd)
        global_rho_qm_clgd = self.cl.gd.zeros(global_array=True)
        local_rho_qm_clgd = self.cl.gd.zeros()

        if self.rank == 0:
            # Coarsen the quantum density
            global_rho_qm_qmgd_coarsened = global_rho_qm_qmgd
            for n in range(self.num_refinements):
                global_rho_qm_qmgd_coarsened = \
                    self.cl.coarseners[n].apply(global_rho_qm_qmgd_coarsened)

            # Add the coarsened quantum density
            o1 = self.index_offset_1
            o2 = self.index_offset_2
            global_rho_qm_clgd[o1[0]:o2[0] - 1,
                               o1[1]:o2[1] - 1,
                               o1[2]:o2[2] - 1] = \
                global_rho_qm_qmgd_coarsened[:]

        # Distribute the combined density to all processes
        self.cl.gd.distribute(global_rho_qm_clgd, local_rho_qm_clgd)

        # Solve quantum potential on classical grid
        local_phi_qm_clgd = self.old_local_phi_qm_clgd
        self.cl.poisson_solver.remove_moment = None
        niter_qm_clgd = self.cl.poisson_solver.solve(phi=local_phi_qm_clgd,
                                                     rho=local_rho_qm_clgd,
                                                     **kwargs)
        self.cl.poisson_solver.remove_moment = self.remove_moment_cl
        self.old_local_phi_qm_clgd = local_phi_qm_clgd.copy()

        # Sum quantum and classical potentials
        local_phi_tot_qmgd, local_phi_tot_clgd = \
            self.couple(qm_qmgd=local_phi_qm_qmgd,
                        qm_clgd=local_phi_qm_clgd,
                        cl_qmgd=local_phi_cl_qmgd,
                        cl_clgd=local_phi_cl_clgd)

        return local_phi_tot_qmgd, local_phi_tot_clgd, \
            (niter_qm, niter_cl, niter_qm_clgd)


class MultipolesPotentialCoupler(PotentialCoupler):
    def __init__(self, *args, **kwargs):
        # The latest appearance of this class is in
        # trac.fysik.dtu.dk/projects/gpaw/browser/
        # branches/electrodynamics/gpaw/fdtd/potential_couplers.py?rev=11708
        #
        # Re-added here a dummy class for compatibility with FDTDPoissonSolver.
        # See above code if you want to reintroduce the working code here.
        raise RuntimeError("MultipolesPotentialCoupler not supported." +
                           "Check the code.")



import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.utilities import AtomsTooClose


def test_atoms_too_close():
    atoms = Atoms('H2', [(0.0, 0.0, 0.0),
                         (0.0, 0.0, 3.995)],
                  cell=(4, 4, 4), pbc=True)

    calc = GPAW(txt=None)
    atoms.calc = calc

    with pytest.raises(AtomsTooClose):
        atoms.get_potential_energy()



import ase.db
from ase.units import kcal, mol
from gpaw.data.g2_1_ref import ex_atomization, atomization


c = ase.db.connect('results.db')

# Energy of atoms:
atoms = {}
for d in c.select(natoms=1):
    atoms[d.numbers[0]] = [d.energy, d.exx]

maepbe = 0.0
maeexx = 0.0
print('                 PBE                   EXX')
print(('-' * 48))
for d in c.select('natoms>1'):
    epberef = atomization[d.name][2] * kcal / mol
    eexxref = ex_atomization[d.name][0] * kcal / mol

    epbe = sum(atoms[atom][0] for atom in d.numbers) - d.energy
    eexx = sum(atoms[atom][1] for atom in d.numbers) - d.exx

    maepbe += abs(epbe - epberef) / len(ex_atomization)
    maeexx += abs(eexx - eexxref) / len(ex_atomization)

    print(('%-4s %10.3f %10.3f %10.3f %10.3f' %
          (d.name, epbe, epbe - epberef, eexx, eexx - eexxref)))

print(('-' * 48))
print(('MAE  %10.3f %10.3f %10.3f %10.3f' % (0, maepbe, 0, maeexx)))

assert maepbe < 0.025
assert maeexx < 0.05



import pytest
from ase.build import molecule

from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.analyse.overlap import Overlap
import gpaw.solvation as solv
from gpaw.lrtddft import LrTDDFT
from gpaw.poisson import PoissonSolver


@pytest.mark.skip(reason='TODO')
def test_solvation_overlap():
    """Check whether LrTDDFT in solvation works"""

    h = 0.4
    box = 2
    nbands = 2

    H2 = Cluster(molecule('H2'))
    H2.minimal_box(box, h)

    c1 = GPAW(h=h, txt=None, nbands=nbands)
    c1.calculate(H2)

    c2 = solv.SolvationGPAW(h=h,
                            txt=None,
                            nbands=nbands + 1,
                            **solv.get_HW14_water_kwargs())
    c2.calculate(H2)
    for poiss in [None, PoissonSolver(nn=c2.hamiltonian.poisson.nn)]:
        lr = LrTDDFT(c2, poisson=poiss)
        print(lr)
    print(Overlap(c1).pseudo(c2))



import xml.sax

import numpy as np

from gpaw.setup_data import search_for_file
from gpaw.atom.radialgd import radial_grid_descriptor


_basis_letter2number = {'s': 1, 'd': 2, 't': 3, 'q': 4}
_basis_number2letter = 'Xsdtq56789'


def parse_basis_name(name):
    """Parse any basis type identifier: 'sz', 'dzp', 'qztp', '4z3p', ... """

    zetacount = _basis_letter2number.get(name[0])
    if zetacount is None:
        zetacount = int(name[0])
    assert name[1] == 'z'

    if len(name) == 2:
        polcount = 0
    elif len(name) == 3:
        assert name[-1] == 'p'
        polcount = 1
    else:
        assert len(name) == 4 and name[-1] == 'p'
        polcount = _basis_letter2number.get(name[2])
        if polcount is None:
            polcount = int(name[2])

    return zetacount, polcount


def get_basis_name(zetacount, polarizationcount):
    zetachar = _basis_number2letter[zetacount]
    if polarizationcount == 0:
        return '%sz' % zetachar
    elif polarizationcount == 1:
        return '%szp' % zetachar
    else:
        polarizationchar = _basis_number2letter[polarizationcount]
        return '%sz%sp' % (zetachar, polarizationchar)


class Basis:
    def __init__(self, symbol, name, readxml=True, rgd=None, world=None):
        self.symbol = symbol
        self.name = name
        self.rgd = rgd
        self.bf_j = []
        self.ribf_j = []
        self.generatorattrs = {}
        self.generatordata = ''
        self.filename = None

        if readxml:
            self.read_xml(world=world)

    @property
    def nao(self):  # implement as a property so we don't have to
        # catch all the places where Basis objects are modified without
        # updating it.  (we can do that later)
        return sum([2 * bf.l + 1 for bf in self.bf_j])

    @property
    def nrio(self):
        return sum([2 * ribf.l + 1 for ribf in self.ribf_j])

    def append(self, bf):
        if bf.type != 'auxiliary':
            self.bf_j.append(bf)
        else:
            self.ribf_j.append(bf)

    def get_grid_descriptor(self):
        return self.rgd

    def tosplines(self):
        return [self.rgd.spline(bf.phit_g, bf.rc, bf.l, points=400)
                for bf in self.bf_j]

    def ritosplines(self):
        return [self.rgd.spline(ribf.phit_g, ribf.rc, ribf.l, points=400)
                for ribf in self.ribf_j]
   
    def read_xml(self, filename=None, world=None):
        parser = BasisSetXMLParser(self)
        parser.parse(filename, world=world)

    def write_xml(self):
        """Write basis functions to file.

        Writes all basis functions in the given list of basis functions
        to the file "<symbol>.<name>.basis".
        """
        if self.name is None:
            filename = '%s.basis' % self.symbol
        else:
            filename = '%s.%s.basis' % (self.symbol, self.name)

        with open(filename, 'w') as fd:
            self.write_to(fd)

    def write_to(self, fd):
        write = fd.write
        write('<paw_basis version="0.1">\n')

        generatorattrs = ' '.join(['%s="%s"' % (key, value)
                                   for key, value
                                   in self.generatorattrs.items()])
        write('  <generator %s>' % generatorattrs)
        for line in self.generatordata.split('\n'):
            write('\n    ' + line)
        write('\n  </generator>\n')

        write('  ' + self.rgd.xml())
       
        # Write both the basis functions and auxiliary ones
        for bfs in [self.bf_j, self.ribf_j]:
            for bf in bfs:
                write(bf.xml(indentation='  '))
        
        write('</paw_basis>\n')

    def reduce(self, name):
        """Reduce the number of basis functions.

        Example: basis.reduce('sz') will remove all non single-zeta
        and polarization functions."""

        zeta, pol = parse_basis_name(name)
        newbf_j = []
        N = {}
        p = 0
        for bf in self.bf_j:
            if 'polarization' in bf.type:
                if p < pol:
                    newbf_j.append(bf)
                    p += 1
            else:
                nl = (int(bf.type[0]), 'spdf'.index(bf.type[1]))
                if nl not in N:
                    N[nl] = 0
                if N[nl] < zeta:
                    newbf_j.append(bf)
                    N[nl] += 1
        self.bf_j = newbf_j

    def get_description(self):
        title = 'LCAO basis set for %s:' % self.symbol
        if self.name is not None:
            name = 'Name: ' + self.name
        else:
            name = 'This basis set does not have a name'
        if self.filename is None:
            fileinfo = 'This basis set was not loaded from a file'
        else:
            fileinfo = 'File: ' + self.filename
        nj = len(self.bf_j)
        count1 = 'Number of radial functions: %d' % nj
        count2 = 'Number of spherical harmonics: %d' % self.nao

        bf_lines = []
        for bf in self.bf_j:
            line = '  l=%d, rc=%.4f Bohr: %s' % (bf.l, bf.rc, bf.type)
            bf_lines.append(line)

        lines = [title, name, fileinfo, count1, count2]
        lines.extend(bf_lines)
        lines.append(f'Number of RI-basis functions {self.nrio}')
        for ribf in self.ribf_j:
            lines.append('l=%d %s' % (ribf.l, ribf.type))
        
        return '\n  '.join(lines)


class BasisFunction:
    """Encapsulates various basis function data."""
    def __init__(self, n=None, l=None, rc=None, phit_g=None, type=''):
        self.n = n
        self.l = l
        self.rc = rc
        self.phit_g = phit_g
        self.type = type
        if n is None or n < 0:
            self.name = 'l=%d %s' % (l, type)
        else:
            self.name = '%d%s %s' % (n, 'spdf'[l], type)

    def __repr__(self, gridid=None):
        txt = '<basis_function '
        if self.n is not None:
            txt += 'n="%d" ' % self.n
        txt += ('l="%r" rc="%r" type="%s"' % (self.l, self.rc, self.type))
        if gridid is not None:
            txt += ' grid="%s"' % gridid
        return txt + '>'

    def xml(self, gridid='grid1', indentation=''):
        txt = indentation + self.__repr__(gridid) + '\n'
        txt += indentation + '  ' + ' '.join(repr(x) for x in self.phit_g)
        txt += '\n' + indentation + '</basis_function>\n'
        return txt


class BasisSetXMLParser(xml.sax.handler.ContentHandler):
    def __init__(self, basis):
        xml.sax.handler.ContentHandler.__init__(self)
        self.basis = basis
        self.type = None
        self.rc = None
        self.data = None
        self.l = None

    def parse(self, filename=None, world=None):
        """Read from symbol.name.basis file.

        Example of filename: N.dzp.basis.  Use sz(dzp) to read
        the sz-part from the N.dzp.basis file."""

        basis = self.basis
        if '(' in basis.name:
            assert basis.name.endswith(')')
            reduced, name = basis.name.split('(')
            name = name[:-1]
        else:
            name = basis.name
            reduced = None
        fullname = '%s.%s.basis' % (basis.symbol, name)
        if filename is None:
            basis.filename, source = search_for_file(fullname, world=world)
        else:
            basis.filename = filename
            with open(filename, 'rb') as fd:
                source = fd.read()

        self.data = None
        xml.sax.parseString(source, self)

        if reduced:
            basis.reduce(reduced)

    def startElement(self, name, attrs):
        basis = self.basis
        if name == 'paw_basis':
            basis.version = attrs['version']
        elif name == 'generator':
            basis.generatorattrs = dict(attrs)
            self.data = []
        elif name == 'radial_grid':
            basis.rgd = radial_grid_descriptor(**attrs)
        elif name == 'basis_function':
            self.l = int(attrs['l'])
            self.rc = float(attrs['rc'])
            self.type = attrs.get('type')
            self.data = []
            if 'n' in attrs:
                self.n = int(attrs['n'])
            elif self.type[0].isdigit():
                self.n = int(self.type[0])
            else:
                self.n = None

    def characters(self, data):
        if self.data is not None:
            self.data.append(data)

    def endElement(self, name):
        basis = self.basis
        if name == 'basis_function':
            phit_g = np.array([float(x) for x in ''.join(self.data).split()])
            bf = BasisFunction(self.n, self.l, self.rc, phit_g, self.type)
            # Also auxiliary basis functions are added here. They are
            # distinguished by their type='auxiliary'.
            basis.append(bf)
        elif name == 'generator':
            basis.generatordata = ''.join([line for line in self.data])


class BasisPlotter:
    def __init__(self, premultiply=True, normalize=False,
                 show=False, save=False, ext='png'):
        self.premultiply = premultiply
        self.show = show
        self.save = save
        self.ext = ext
        self.default_filename = '%(symbol)s.%(name)s.' + ext

        self.title = 'Basis functions: %(symbol)s %(name)s'
        self.xlabel = r'r [Bohr]'
        if premultiply:
            ylabel = r'$\tilde{\phi} r$'
        else:
            ylabel = r'$\tilde{\phi}$'
        self.ylabel = ylabel

        self.normalize = normalize

    def plot(self, basis, filename=None, **plot_args):
        import matplotlib.pyplot as plt
        if plot_args is None:
            plot_args = {}
        r_g = basis.rgd.r_g

        print('Element  :', basis.symbol)
        print('Name     :', basis.name)
        print()
        print('Basis functions')
        print('---------------')

        norm_j = []
        for j, bf in enumerate(basis.bf_j):
            ng = len(bf.phit_g)
            rphit_g = r_g[:ng] * bf.phit_g
            norm = (rphit_g**2 * basis.rgd.dr_g[:ng]).sum()
            norm_j.append(norm)
            print(bf.type, '[norm=%0.4f]' % norm)

        print()
        print('Generator')
        for key, item in basis.generatorattrs.items():
            print('   ', key, ':', item)
        print()
        print('Generator data')
        print(basis.generatordata)

        if self.premultiply:
            factor = r_g
        else:
            factor = np.ones_like(r_g)

        dashes_l = [(), (6, 3), (4, 1, 1, 1), (1, 1)]

        plt.figure()
        for norm, bf in zip(norm_j, basis.bf_j):
            ng = len(bf.phit_g)
            y_g = bf.phit_g * factor[:ng]
            if self.normalize:
                y_g /= norm
            plt.plot(r_g[:ng], y_g, label=bf.type[:12],
                     dashes=dashes_l[bf.l], lw=2,
                     **plot_args)
        axis = plt.axis()
        rc = max([bf.rc for bf in basis.bf_j])
        newaxis = [0., rc, axis[2], axis[3]]
        plt.axis(newaxis)
        plt.legend()
        plt.title(self.title % basis.__dict__)
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)

        if filename is None:
            filename = self.default_filename
        if self.save:
            plt.savefig(filename % basis.__dict__)

        if self.show:
            plt.show()



#!/usr/bin/env python3
"""Bash completion for ase.

Put this in your .bashrc::

    complete -o default -C /path/to/gpaw/cli/complete.py gpaw

or run::

    $ gpaw completion [-0] [<your-bash-file>]

"""

import os
import sys
from glob import glob


def match(word, *suffixes):
    return [w for w in glob(word + '*')
            if any(w.endswith(suffix) for suffix in suffixes)]


# Beginning of computer generated data:
commands = {
    'atom':
        ['-f', '--xc-functional', '-a', '--add', '--spin-polarized', '-d',
         '--dirac', '-p', '--plot', '-e', '--exponents', '-l',
         '--logarithmic-derivatives', '-n', '--ngrid', '-R',
         '--rcut', '-r', '--refine', '-s',
         '--scalar-relativistic', '--no-ee-interaction'],
    'completion':
        [],
    'dataset':
        ['-f', '--xc-functional', '-C', '--configuration', '-P',
         '--projectors', '-r', '--radius', '-0',
         '--zero-potential', '-c',
         '--pseudo-core-density-radius', '-z', '--pseudize',
         '-p', '--plot', '-l', '--logarithmic-derivatives', '-w',
         '--write', '-s', '--scalar-relativistic', '-n',
         '--no-check', '-t', '--tag', '-a', '--alpha', '-g',
         '--gamma', '-b', '--create-basis-set', '--nlcc',
         '--core-hole', '-e', '--electrons', '--ri', '--omega'],
    'diag':
        ['-b', '--bands', '-s', '--scalapack'],
    'dos':
        ['-p', '--plot', '-i', '--integrated', '-w', '--width', '-a',
         '--atom', '-t', '--total', '-r', '--range', '-n',
         '--points', '--soc'],
    'gpw':
        ['-w', '--remove-wave-functions'],
    'info':
        [],
    'install-data':
        ['--version', '--tarball', '--list-all', '--gpaw', '--sg15',
         '--basis', '--test', '--register', '--no-register'],
    'python':
        ['--dry-run', '-z', '-d', '--debug', '--command', '-c',
         '--module', '-m'],
    'run':
        ['-p', '--parameters', '-t', '--tag', '--properties', '-f',
         '--maximum-force', '--constrain-tags', '-s',
         '--maximum-stress', '-E', '--equation-of-state',
         '--eos-type', '-o', '--output', '--modify', '--after',
         '--dry-run', '-w', '--write', '-W', '--write-all'],
    'sbatch':
        ['-0', '--test'],
    'symmetry':
        ['-t', '--tolerance', '-k', '--k-points', '-v', '--verbose', '-s',
         '--symmorphic'],
    'test':
        []}
# End of computer generated data


def complete(word, previous, line, point):
    for w in line[:point - len(word)].strip().split()[1:]:
        if w[0].isalpha():
            if w in commands:
                command = w
                break
    else:
        if word[:1] == '-':
            return ['-h', '--help', '--version', '-P', '--parallel']
        return list(commands.keys()) + ['-h', '--help', '--version',
                                        '-P', '--parallel']

    if word[:1] == '-':
        return commands[command]

    words = []

    if command == 'help':
        words = commands

    elif command == 'test':
        from glob import glob
        path = __file__.rsplit('/', 1)[0] + '/../test/**/*.py'
        return (word.rsplit('/test/')[-1]
                for word in glob(path, recursive=True))

    return words


if __name__ == '__main__':
    word, previous = sys.argv[2:]
    line = os.environ['COMP_LINE']
    point = int(os.environ['COMP_POINT'])
    words = complete(word, previous, line, point)
    for w in words:
        if w.startswith(word):
            print(w)



from gpaw.basis_data import BasisFunction
from collections import defaultdict


def generate_ri_basis(basis, accuracy):
    lmax = 2

    # TODO: Hartree
    def poisson(n_g, l):
        return Hartree(basis.rgd, n_g, l)

    # Auxiliary basis functions per angular momentum channel
    auxt_lng = defaultdict(lambda: [])

    def add(aux_g, l, rc=None):
        ribf = BasisFunction(n=None, l=l, rc=rc, phit_g=aux_g,
                             type='auxiliary')
        basis.append(ribf)

    def basisloop():
        for j, bf in enumerate(basis.bf_j):
            yield j, bf.l, bf.rc, bf.phit_g

    # Double basis function loop to create product orbitals
    for j1, l1, rc1, phit1_g in basisloop():
        for j2, l2, rc2, phit2_g in basisloop():
            # Loop only over ordered pairs
            if j1 > j2:
                continue

            # Loop over all possible angular momentum states what the
            # product l1 x l2 creates.
            for l in range((l1 + l2) % 2, l1 + l2 + 1, 2):
                if l > lmax:
                    continue
 
                # min: The support of basis function is the intersection
                # of the individual supports.
                add(phit1_g * phit2_g, l, rc=min(rc1, rc2))

    for l, auxt_ng in auxt_lng.items():
        print(l, auxt_ng)
        print(f'    l={l}')
        for n, auxt_g in enumerate(auxt_ng):
            print(f'        {n}')


def Hartree(rgd, n_g, l):
    v_g = rgd.poisson(n_g, l)
    v_g[1:] /= rgd.r_g[1:]
    v_g[0] = v_g[1]
    return v_g



from gpaw.analyse.vdwradii import vdWradii

# data from A. Bondi, J. Phys. Chem. 68 (1964) 441


def test_vdwradii():
    data_Bondi = {  # units Anstrom
        'He': 1.40,
        'Ne': 1.54,
        'Ar': 1.88,
        'Kr': 2.02,
        'Xe': 2.16
    }
    # data from Felix Hanke (FHI-AIMS ?)
    data_Hanke = {  # units Anstrom
        'H': 1.640449351,
        'C': 1.8997461838999998,
        'N': 1.7674518813999998,
        'O': 1.6880752998999997,
        'Cu': 1.9897063095999996,
    }
    for symbol in ['He', 'Ne', 'Ar', 'Kr', 'H', 'C', 'N', 'O', 'Cu']:
        R = vdWradii([symbol], 'PBE')[0]
        if symbol in data_Bondi:
            Rref = data_Bondi[symbol]
        else:
            Rref = data_Hanke[symbol]
        error = abs(R - Rref)
        print("symbol, R, Rref, error:", symbol, R, Rref, error)
        assert error < 0.05



""" This module implements absorbing boundary conditions to minimize errors
caused by boundary reflections. The first classes use negative imaginary
potentials of different shapes which are known to absorb waves. The last
class PML uses the perfectly matched layer approach.

    For information about imaginary potentials please see:

    Takashi Nakatsukasa, Kazuhiro Yabana, J. Chem. Phys. 114, 2550 (2001);
    doi:10.1063/1.1338527
    Daniel Neuhasuer and Michael Baer,  J. Chem. Phys. 90, 4351 (1989);
    doi:10.1063/1.456646

    About PML:
    Chunxiong Zheng, Journal of Computational  Physics,
    volume 227, Issue 1(Nov 2007) pages 537-556;
    doi:10.1016/j.jcp.2007.08.004
"""

import numpy as np

from ase.units import Bohr


class DummyAbsorbingBoundary:
    """ Virtual and not usable class for abosrbing boundaries."""
    def __init__(self):
        self.v_imag = None
        self.type = None

    def set_up(self):
        pass

    def get_potential_matrix(self):
        return self.v_imag


class LinearAbsorbingBoundary(DummyAbsorbingBoundary):
    """
    This class uses linear negative imaginary potential for absorption.
    For distances larger than abc_range, the negative imaginary potential
    increases linearly. Positive float abc_strength describes the steepness
    of the slope. High values of abc_strength cause more reflection from
    the potential and small values do not absorb enough.

    Parameters:
      abc_range: Positive float number. Absorbing potential is zero
                 where distance from the middle point or chosen positions
                 in the grid is smaller and starts to increase linearly
                 after this point.
      abc_strength: Positive float number. A good value for this is usually
                    about 0.01

      positions: An array of positions in the grid. Each of these points
                 has abc_range of free space around them. If this is left
                 as None, the middle point of the grid is used. You can use
                 the position array of an ASE Atoms object here.
    """
    def __init__(self, abc_range, abc_strength, positions=None):
        DummyAbsorbingBoundary.__init__(self)
        self.abc_strength = abc_strength
        self.abc_r = abc_range / Bohr
        self.positions = positions / Bohr
        self.type = 'IPOT'

    def set_up(self, gd):
        """
        Creates the potential matrix self.v_imag.

        Parameters:

        gd: grid descriptor
        """

        assert gd.orthogonal

        # self.v_imag = np.zeros((gd.n_c[0],gd.n_c[1],gd.n_c[2]),dtype=complex)
        self.v_imag = gd.zeros(dtype=complex)

        # If positions array wasn't given, uses the middle point of the grid.
        if self.positions is None:
            self.positions = [
                np.array([
                    gd.N_c[0] * gd.h_cv[0, 0] * 0.5,  # middle
                    gd.N_c[1] * gd.h_cv[1, 1] * 0.5,
                    gd.N_c[2] * gd.h_cv[2, 2] * 0.5
                ])
            ]

        for i in range(gd.n_c[0]):
            x = (i + gd.beg_c[0]) * gd.h_cv[0, 0]
            for j in range(gd.n_c[1]):
                y = (j + gd.beg_c[1]) * gd.h_cv[1, 1]
                for k in range(gd.n_c[2]):
                    z = (k + gd.beg_c[2]) * gd.h_cv[2, 2]
                    position = np.array([x, y, z])
                    # Calculates the distance from the nearest chosen point
                    # in the grid
                    position_vectors = self.positions - position

                    r = np.linalg.norm(position_vectors[0])
                    for vector in position_vectors:
                        if np.linalg.norm(vector) < r:
                            r = np.linalg.norm(vector)

                    if r > self.abc_r:
                        self.v_imag[i][j][k] = \
                            -(0 + 1j) * self.abc_strength * (r - self.abc_r)


class P4AbsorbingBoundary(DummyAbsorbingBoundary):
    """
    The negative imaginary potential used by this class are 4th degree
    polynomials which are constructed so that the value is zero at the
    beginning and abc_strength at the end. The derivative is
    zero at the begininning and zero at the end.

    Parameters:
      abc_range: Positive float number. Absorbing potential is zero where
                 distance from the middle point or chosen positions in the
                 grid is smaller and starts to increase after this point.

      abc_strength: Positive float number.

      positions: An array of positions in the grid. Each of these points has
                 abc_range of free space around them. If this is left as None,
                 the middle point of the grid is used. You can use the position
                 array of an ASE Atoms object here. You have to define the
                 width parameter if you use this.

      width:   The width of the absorbing layer. If you don't define positions
               array a value for this is automatically generated so that the
               width is from abc_range to the end of the grid (works best for
               cube).
               If you use the atom-centered model you have to define the
               width yourself.
    """
    def __init__(self, abc_range, abc_strength, positions=None, width=None):
        DummyAbsorbingBoundary.__init__(self)
        self.abc_r = abc_range / Bohr
        self.abc_strength = abc_strength
        self.positions = positions / Bohr
        self.width = width / Bohr
        self.type = 'IPOT'

    def set_up(self, gd):

        # self.v_imag = np.zeros((gd.n_c[0],gd.n_c[1],gd.n_c[2]),dtype=complex)
        self.v_imag = gd.zeros(dtype=complex)

        # If positions array wasn't given, uses the middle point of the
        # grid as the center.
        if self.positions is None:
            self.positions = [
                np.array([
                    gd.N_c[0] * gd.h_cv[0, 0] * 0.5,  # middle
                    gd.N_c[1] * gd.h_cv[1, 1] * 0.5,
                    gd.N_c[2] * gd.h_cv[2, 2] * 0.5
                ])
            ]

        if self.width is None:
            self.width = np.linalg.norm(
                self.positions[0]) / np.sqrt(3) - self.abc_r

        for i in range(gd.n_c[0]):
            x = (i + gd.beg_c[0]) * gd.h_cv[0, 0]
            for j in range(gd.n_c[1]):
                y = (j + gd.beg_c[1]) * gd.h_cv[1, 1]
                for k in range(gd.n_c[2]):
                    z = (k + gd.beg_c[2]) * gd.h_cv[2, 2]
                    position = np.array([x, y, z])

                    position_vectors = self.positions - position

                    r = np.linalg.norm(position_vectors[0])
                    for vector in position_vectors:
                        if np.linalg.norm(vector) < r:
                            r = np.linalg.norm(vector)

                    if r > self.abc_r:
                        if r < self.abc_r + self.width:
                            self.v_imag[i][j][k] = (
                                (0 + 1j) *
                                ((np.sqrt(self.abc_strength) -
                                  np.sqrt(self.abc_strength) / self.width**2
                                  * (r - self.abc_r)**2)**2 -
                                 self.abc_strength))
                        else:
                            self.v_imag[i][j][k] = -self.abc_strength * (0 +
                                                                         1j)


class PML:
    """
    Important! You must use the BiCGStab solver or your time propagation
    will likely crash. Give the parameter solver='BiCGStab' as you create
    the TDDFT object.

    Using PML makes the time progation slower so reserve twice more time.
    As imaginary potential is usually almost equally good, this is mostly
    for testing (until improved?).

    And now for something completely different, a Perfectly matched layer.
    Note that you can't give positions array as parameter for this class.

    Parameters:
      abc_range:     Starting from the center of the grid, the amount of
                     free space before PML starts to affect.

      abc_strength:   Positive float, good amount should be around 0.1
    """
    def __init__(self, abc_range, abc_strength):

        self.abc_range = abc_range / Bohr
        self.abc_strength = abc_strength
        self.type = 'PML'

    def set_up(self, gd):
        r"""Set up matrices for PML.

        Creates the matrixes needed when the PML is applied in tdopers.py
        when applying time-dependent hamiltonian.

        Perfectly matched layer is applied as potential Vpml = Tpml-T,
        Where  T = -.5*\nabla^{2}\psi  (Use latex for equations)

        Tpml we get from

        'A perfectly matched layer approach to the nonlinear Schrodinger wave
        equations',
        Journal of Computational  Physics,
        volume 227, Issue 1(Nov 2007) pages 537-556,
        Author: Chunxiong Zheng

        T_{pml} = -0.5*(G\nabla G\nabla \psi + G^{2}\nabla^{2}\psi)

        where G = \frac{1}{1+R\sigma}

        V_{pml} = -0.5 * (G\nabla G\nabla \psi + (G^{2}-1)\nabla^{2}\psi)

        This is probably not the most optimal approach and slows the
        propagation.

        The matrixes created here are G and the gradients of G in all
        directions.

        """
        R = (0 + 1j)  # Complex number, has to be in the first quadrant.
        self.G = gd.zeros(dtype=complex)
        self.G[:] = 1.0

        self.dG = gd.zeros(n=3, dtype=complex)

        r0 = np.array([
            gd.N_c[0] * gd.h_cv[0, 0] * 0.5, gd.N_c[1] * gd.h_cv[1, 1] * 0.5,
            gd.N_c[2] * gd.h_cv[2, 2] * 0.5
        ])  # middle point
        for i in range(gd.n_c[0]):
            x = (i + gd.beg_c[0]) * gd.h_cv[0, 0]
            for j in range(gd.n_c[1]):
                y = (j + gd.beg_c[1]) * gd.h_cv[1, 1]
                for k in range(gd.n_c[2]):
                    z = (k + gd.beg_c[2]) * gd.h_cv[2, 2]
                    position = np.array([x, y, z])
                    r = np.linalg.norm(position - r0)

                    if r > self.abc_range:
                        self.G[i][j][k] = (1.0 + R * self.abc_strength *
                                           (r - self.abc_range)**2)**-1.0
                        self.dG[0][i][j][k] = (
                            -(1.0 + R *
                              self.abc_strength *
                              (r - self.abc_range)**2.0)**-2.0 * 2.0 * R *
                            self.abc_strength * (r - self.abc_range) *
                            (x - r0[0]) / r)

                        self.dG[1][i][j][k] = -(
                            1.0 + R * self.abc_strength *
                            (r - self.abc_range)**2.0
                        )**-2.0 * 2.0 * R * self.abc_strength * (
                            r - self.abc_range) * (y - r0[1]) / r

                        self.dG[2][i][j][k] = -(
                            1.0 + R * self.abc_strength *
                            (r - self.abc_range)**2.0
                        )**-2.0 * 2.0 * R * self.abc_strength * (
                            r - self.abc_range) * (z - r0[2]) / r

    def get_G(self):
        return self.G

    def get_dG(self):
        return self.dG

    def isPML(self):
        return True



"""Density functional perturbation theory (DFPT) package.

DFPT
====

In DFPT the first-order variation in the density due to a static perturbation
in the external potential is calculated self-consistently from a set of coupled
equations for:

    1. the variation of the effective potential,
    2. variation of the wave functions, and
    3. the resulting variation of the density.

The variation of the wave-functions can be determined from a set of linear
equations, the so-called Sternheimer equation. Since only their projection onto
the unoccupied state manifold is required to determine the corresponding
density variation, only the occupied wave functions/states of the unperturbed
system are needed. This is in contrast to the standard textbook expression for
the first-order variation of the wave function which involves the full spectrum
of the unpertured system.

The first-order variation of the density with respect to different
perturbations can be used to obtain various physical quantities of the
system. This package includes calculators for:

    1. Phonons in periodic systems
        - perturbation is a lattice distortion with a given q-vector (work in
          progress).
    2. Born effective charges
        - perturbation is a constant electric field (to be done).
    3. Dielectric constant
        - perturbation is a constant electric field (to be done).


References
----------

Publications from various groups giving a good overview of DFPT:

    - Louie et al., Phys. Rev. B 76, 165108 (2007)
    - Vanderbilt et al., Phys. Rev. B 74, 054304 (2006)
    - Clark et al., Phys. Rev. B 73, 155114 (2006)
    - Baroni et al., Rev. Mod. Phys. 73, 515 (2001)
    - Gonze, Phys. Rev. B 55, 10337 (1997)
    - Gonze et al., Phys. Rev. B 55, 10355 (1997)
    - Savrasov, Phys. Rev. B 54, 16470 (1996)
    - Savrasov et al., Phys. Rev. B 54, 16487 (1996)

Implementaion notes
===================

1. When using the ``derivative`` method of the ``lfc`` class to calculate
   integrals between derivatives of localized functions wrt atomic
   displacements and functions defined on the grid, the signs of the calculated
   integrals come out incorrectly (see doc string of the ``derivative``
   method). All files in this ``dfpt`` package follows the prescription::

        X_niv = -1 * X_aniv[a] ,

   for correcting the sign error, i.e. correct the sign when extracting from
   the dictionary.

2. Arrays containing first-order derivatives should be suffixed with a ``1``
   in the variable name, e.g. ``nt1_g`` for the density.

3. Naming conventions for arrays:

    * q-vectors for the dynamical matrix:
        - q: index in irreducible BZ.
        - k: index in the full BZ.

    * atomic and cartesian indices in the dynamical matrix
        - av: be careful, can both refer to a composite atomic and cartesian
              index (i.e. corresponds only to ONE dimension in the array) and
              individual atomic and cartesian indices (corresponding to TWO
              dimensions in an array).

"""

# flake8: noqa
# __version__ = "0.1"

# Import things specified in the __all__ attributes
from gpaw.dfpt.phononcalculator import *
from gpaw.dfpt.responsecalculator import *
from gpaw.dfpt.phononperturbation import *
from gpaw.dfpt.dynamicalmatrix import *
from gpaw.dfpt.electronphononcoupling import *

# Set the behavior of from gpaw.dfpt import *
import gpaw.dfpt.phononcalculator
import gpaw.dfpt.responsecalculator
import gpaw.dfpt.phononperturbation
import gpaw.dfpt.dynamicalmatrix
import gpaw.dfpt.electronphononcoupling



"""Methods to calculate material properties in the Heisenberg model.
Primarily based on the magnon dispersion relations."""

import numpy as np


def calculate_single_site_magnon_energies(J_qx, q_qc, mm):
    """Compute the magnon energies from the isotropic exchange constants of a
    system with a single magnetic site in the unit cell, as a function of the
    wave vector q:

    ħω(q) = g μ_B / M [J(0) - J(q)]

    Parameters
    ----------
    J_qx : np.ndarray
        Isotropic exchange constants as a function of q.
        J_qx can have any number of additional dimensions x, which will be
        treated independently.
    q_qc : np.ndarray
        q-vectors in relative coordinates. Has to include q=0.
    mm : float
        Magnetic moment of the site in μ_B.

    Returns
    -------
    E_qx : np.ndarray
        Magnon energies as a function of q and x. Same shape as input J_qx.
    """
    assert J_qx.shape[0] == q_qc.shape[0]
    assert np.allclose(J_qx.imag, 0.),\
        "The exchange constants of an isotropic system with a single "\
        "magnetic sublattice are always real."

    q0 = get_q0_index(q_qc)
    J0_x = J_qx[q0]

    # Compute energies
    E_qx = 2. / mm * (J0_x[np.newaxis, ...] - J_qx)

    return E_qx.real


def calculate_fm_magnon_energies(J_qabx, q_qc, mm_ax):
    """Compute the magnon eigenmode energies from the isotropic exchange
    constants of a ferromagnetic system with an arbitrary number of magnetic
    sites in the unit cell, as a function of the wave vector q.

    The eigenmodes are calculated as the eigenvalues to the dynamic spin wave
    matrix:

    H^ab(q) = g μ_B / sqrt(M_a M_b) [Σ_c J^ac(0) δ_ab - J^ab(q)]

    Parameters
    ----------
    J_qabx : np.ndarray
        Isotropic exchange constants as a function of q and sublattice indices
        a and b. J_qabx can have any number of additional dimensions x, which
        will be treated independently.
    q_qc : np.ndarray
        q-vectors in relative coordinates. Has to include q=0.
    mm_ax : np.ndarray
        Magnetic moments of the sublattice sites in μ_B.

    Returns
    -------
    E_qnx : np.ndarray
        Magnon eigenmode energies as a function of q, mode index n and x.
    """
    H_qabx = generate_fm_dynamic_spin_wave_matrix(J_qabx, q_qc, mm_ax)

    # Move magnetic site axes in order to prepare for np.linalg.eig
    H_qbxa = np.moveaxis(H_qabx, 1, -1)
    H_qxab = np.moveaxis(H_qbxa, 1, -1)

    # Diagonalize the matrix
    E_qxn, _ = np.linalg.eig(H_qxab)

    # Transpose to output format
    E_qnx = np.moveaxis(E_qxn, -1, 1)

    # Eigenvalues should be real, otherwise input J_qabx has been invalid
    assert np.allclose(E_qnx.imag, 0.)
    E_qnx = E_qnx.real

    return E_qnx


def generate_fm_dynamic_spin_wave_matrix(J_qabx, q_qc, mm_ax):
    """Generate the dynamic spin wave matrix from the isotropic exchange
    constants of a ferromagnet:

    H^ab(q) = g μ_B / sqrt(M_a M_b) [Σ_c J^ac(0) δ_ab - J^ab(q)]

    Parameters
    ----------
    J_qabx : np.ndarray
        Isotropic exchange constants as a function of q and sublattice indices
        a and b. J_qabx can have any number of additional dimensions x, which
        will be treated independently.
    q_qc : np.ndarray
        q-vectors in relative coordinates. Has to include q=0.
    mm_ax : np.ndarray
        Magnetic moments of the sublattice sites in μ_B.

    Returns
    -------
    H_qabx : np.ndarray
        Dynamic spin wave matrix. Has the same shape as the input J_qabx
    """
    assert len(J_qabx.shape) >= 3
    assert J_qabx.shape[1] == J_qabx.shape[2]
    assert J_qabx.shape[1] == mm_ax.shape[0]
    assert J_qabx.shape[0] == q_qc.shape[0]
    assert J_qabx.shape[3:] == mm_ax.shape[1:]
    assert np.allclose(J_qabx, np.conj(np.moveaxis(J_qabx, 2, 1))),\
        "The isotropic exchange constants J^ab(q) are Hermitian by definition."
    na = mm_ax.shape[0]

    # Get J^ab(0)
    q0 = get_q0_index(q_qc)
    J0_acx = J_qabx[q0]
    assert np.allclose(J0_acx.imag, 0.),\
        "For a collinear system without spin-orbit coupling, the exchange "\
        "constants are not only isotropic, but also reciprocal. In "\
        "particular, this implies that [J^ab(q)]^*=J^ab(-q)."

    # Set up magnetic moment prefactor as outer product
    mm_inv_abx = 2. / np.sqrt(mm_ax[:, np.newaxis, ...]
                              * mm_ax[np.newaxis, ...])

    # Calculate diagonal component Σ_c J^ac(0) δ_ab
    J0_ax = np.sum(J0_acx, axis=1)
    diagonal_mapping = np.zeros((na, na, na))
    np.fill_diagonal(diagonal_mapping, 1)
    J0_abx = np.tensordot(diagonal_mapping, J0_ax, axes=(2, 0))

    # Calculate the dynamic spin wave matrix
    H_qabx = mm_inv_abx[np.newaxis, ...] * (J0_abx[np.newaxis, ...] -
                                            J_qabx)

    return H_qabx


def get_q0_index(q_qc):
    """Find index corresponding q=0 in q-vector array."""
    q0_indices = np.argwhere(np.all(q_qc == 0, axis=1))

    if len(q0_indices) >= 1:
        return int(q0_indices[0])
    else:
        raise ValueError('q_qc has to include q=0, i.e. q_c = [0., 0., 0.]')



r"""Finite difference calculation of changes in potential.

The implementation is based on finite-difference calculations of the the atomic
gradients of the effective potential expressed on a real-space grid. The el-ph
couplings are obtained from LCAO representations of the atomic gradients of the
effective potential and the electronic states.

In PAW the matrix elements of the derivative of the effective potential is
given by the sum of the following contributions::

                  d                  d
            < i | -- V | j > = < i | -- V | j>
                  du  eff            du

                               _
                              \        ~a     d   .       ~a
                            +  ) < i | p  >   -- /_\H   < p | j >
                              /_        i     du     ij    j
                              a,ij

                               _
                              \        d  ~a     .        ~a
                            +  ) < i | -- p  >  /_\H    < p | j >
                              /_       du  i        ij     j
                              a,ij

                               _
                              \        ~a     .        d  ~a
                            +  ) < i | p  >  /_\H    < -- p  | j >
                              /_        i        ij    du  j
                              a,ij

where the first term is the derivative of the potential (Hartree + XC) and the
last three terms originate from the PAW (pseudopotential) part of the effective
DFT Hamiltonian.

Note: It is a bit difficult to find good references for how spin-polarisation
is supposed to be handled. Here we just handle the spin channels separately.
Use with care.

"""
import numpy as np

from ase import Atoms
from ase.phonons import Displacement

from gpaw.calculator import GPAW

dr_version = 1
# v1: saves natom, supercell, delta


class DisplacementRunner(Displacement):
    """Class for calculating the changes in effective potential.

    The derivative of the effective potential wrt atomic displacements is
    obtained from a finite difference approximation to the derivative by doing
    a self-consistent calculation for atomic displacements in the +/-
    directions. These calculations are carried out in the ``run`` member
    function.
    """

    def __init__(self, atoms: Atoms, calc: GPAW, supercell: tuple = (1, 1, 1),
                 name: str = "elph", delta: float = 0.01,
                 calculate_forces: bool = True) -> None:
        """Initialize with base class args and kwargs.

        Parameters
        ----------
        atoms: Atoms
            The atoms to work on. Primitive cell.
        calc: GPAW
            Calculator for the supercell finite displacement calculation.
        supercell: tuple, list
            Size of supercell given by the number of repetitions (l, m, n) of
            the small unit cell in each direction.
        name: str
            Name to use for files (default: 'elph').
        delta: float
            Magnitude of displacements. (default: 0.01 A)
        calculate_forces: bool
            If true, also calculate and store the dynamical matrix.
        """

        # Init base class and make the center cell in the supercell the
        # reference cell
        Displacement.__init__(self, atoms, calc=calc, supercell=supercell,
                              name=name, delta=delta, center_refcell=True)
        self.calculate_forces = calculate_forces

    def calculate(self, atoms_N: Atoms, disp):
        return self(atoms_N)

    def __call__(self, atoms_N: Atoms) -> dict:
        """Extract effective potential and projector coefficients."""

        # Do calculation
        atoms_N.get_potential_energy()

        # Calculate forces if desired
        if self.calculate_forces:
            forces = atoms_N.get_forces()
        else:
            forces = None

        # Get calculator
        calc = atoms_N.calc
        if not isinstance(calc, GPAW):
            calc = calc.dft  # unwrap DFTD3 wrapper

        # Effective potential (in Hartree) and projector coefficients
        # Note: Need to use coarse grid, because we project onto basis later
        Vt_sG = calc.hamiltonian.vt_sG
        Vt_sG = calc.wfs.gd.collect(Vt_sG, broadcast=True)
        dH_asp = calc.hamiltonian.dH_asp

        setups = calc.wfs.setups
        nspins = calc.wfs.nspins

        dH_all_asp = {}
        for a, setup in enumerate(setups):
            ni = setup.ni
            nii = ni * (ni + 1) // 2
            dH_tmp_sp = np.zeros((nspins, nii))
            if a in dH_asp:
                dH_tmp_sp[:] = dH_asp[a]
            calc.wfs.gd.comm.sum(dH_tmp_sp)
            dH_all_asp[a] = dH_tmp_sp

        output = {"Vt_sG": Vt_sG, "dH_all_asp": dH_all_asp}
        if forces is not None:
            output["forces"] = forces
        return output

    def save_info(self) -> None:
        with self.cache.lock("info") as handle:
            if handle is not None:
                info = {"natom": len(self.atoms), "supercell": self.supercell,
                        "delta": self.delta, "dr_version": dr_version}
                handle.save(info)

    def run(self) -> None:
        """Run the calculations for the required displacements."""
        # Save some information about this run
        self.save_info()
        Displacement.run(self)



from types import SimpleNamespace

from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.lfc import BasisFunctions
from gpaw.mpi import serial_comm


def create_basis(ibz,
                 nspins,
                 pbc_c,
                 grid,
                 setups,
                 dtype,
                 fracpos_ac,
                 world=serial_comm,
                 kpt_comm=serial_comm):
    kd = KPointDescriptor(ibz.bz.kpt_Kc, nspins)
    kd.set_symmetry(SimpleNamespace(pbc=pbc_c),
                    ibz.symmetries.symmetry,
                    comm=world)
    kd.set_communicator(kpt_comm)

    basis = BasisFunctions(grid._gd,
                           [setup.basis_functions_J for setup in setups],
                           kd,
                           dtype=dtype,
                           cut=True)
    basis.set_positions(fracpos_ac)
    return basis



import numpy as np
import pytest

from ase.lattice.hexagonal import Graphene
from ase.parallel import parprint as pp

from gpaw import GPAW
from gpaw.response.df import DielectricFunction
from gpaw.mpi import world


@pytest.mark.response
@pytest.mark.skip(reason='TODO')
def test_graphene_EELS():
    system = Graphene(symbol='C',
                      latticeconstant={'a': 2.45, 'c': 1.0},
                      size=(1, 1, 1))
    system.pbc = (1, 1, 0)
    system.center(axis=2, vacuum=4.0)

    nkpts = 5

    communicator = world.new_communicator(np.array([world.rank]))
    gpwname = 'dump.graphene.gpw'

    if world.rank == 0:
        calc = GPAW(mode='pw',
                    kpts=(nkpts, nkpts, 1),
                    communicator=communicator,
                    xc='oldLDA',
                    nbands=len(system) * 6,
                    txt='gpaw.graphene.txt')
        system.calc = calc
        system.get_potential_energy()
        calc.write(gpwname, mode='all')

    world.barrier()

    parallel = dict(domain=(1, 1, 1), band=1)
    if world.size == 8:
        # parallel['domain'] = (1, 1, 2)
        parallel['band'] = 2
    calc = GPAW(gpwname,
                txt=None,
                parallel=parallel)
    pp('after restart')

    q = np.array([1.0 / nkpts, 0., 0.])
    w = np.linspace(0, 31.9, 320)
    dw = w[1] - w[0]

    def getpeak(energies, loss):
        arg = loss.argmax()
        energy = energies[arg]
        peakloss = loss[arg]
        return energy, peakloss

    scriptlines = []

    loss_errs = []
    energy_errs = []

    def check(name, energy, peakloss, ref_energy, ref_loss):
        pp('check %s :: energy = %5.2f [%5.2f], peakloss = %.12f [%.12f]'
           % (name, energy, ref_energy, peakloss, ref_loss))
        energy_errs.append(abs(energy - ref_energy))
        loss_errs.append(abs(peakloss - ref_loss))

    template = """\
    check_df('%s', %s, %s, %s, %s,
             **%s)"""

    def check_df(name, ref_energy, ref_loss, ref_energy_lfe, ref_loss_lfe,
                 **kwargs_override):
        kwargs = dict(calc=calc, frequencies=w.copy(), eta=0.5, ecut=30,
                      txt='df.%s.txt' % name)
        kwargs.update(kwargs_override)
        df = DielectricFunction(**kwargs)
        fname = 'dfdump.%s.dat' % name
        df.get_eels_spectrum('RPA', q_c=q, filename=fname)
        world.barrier()
        d = np.loadtxt(fname, delimiter=',')

        loss = d[:, 1]
        loss_lfe = d[:, 2]
        energies = d[:, 0]

        # import pylab as pl
        # fig = pl.figure()
        # ax1 = fig.add_subplot(111)
        # ax1.plot(d[:, 0], d[:, 1]/np.max(d[:, 1]))
        # ax1.plot(d[:, 0], d[:, 2]/np.max(d[:, 2]))
        # ax1.axis(ymin=0, ymax=1)
        # fig.savefig('fig.%s.pdf' % name)

        energy, peakloss = getpeak(energies, loss)
        energy_lfe, peakloss_lfe = getpeak(energies, loss_lfe)

        check(name, energy, peakloss, ref_energy, ref_loss)
        check('%s-lfe' % name, energy_lfe, peakloss_lfe, ref_energy_lfe,
              ref_loss_lfe)

        line = template % (name, energy, peakloss, energy_lfe, peakloss_lfe,
                           repr(kwargs_override))
        scriptlines.append(line)

    # These lines can be generated by the loop over 'scriptlines' below,
    # in case new reference values are wanted

    # The implementation for vcut choices has still to be done

    check_df('3d', 20.20, 2.505295593820, 26.90, 1.748517033160)
    #         **{'rpad': array([1, 1, 1])}) #, 'vcut': '3D'})
    # check_df('2d', 20.10, 2.449662058530, 26.80, 1.538080502420,
    #          **{'rpad': array([1, 1, 1]), 'vcut': '2D'})

    pp()
    pp('Insert lines into script to set new reference values:')
    for line in scriptlines:
        pp(line)
    pp()

    for err in energy_errs:
        # with the current grid this error just means
        assert err < dw / 4.0, err
    for err in loss_errs:
        assert err < 1e-6, err



import numpy as np
import pytest

from ase.build import bulk
from ase.parallel import world
from gpaw import GPAW, FermiDirac, PW


def get_calculator(sl_auto, kpoint_gamma):
    calculator = GPAW(
        mode=PW(100),
        occupations=FermiDirac(0.1),
        parallel={'sl_auto': sl_auto, 'band': 2 if world.size == 8 else 1},
        kpts=[1, 1, 1] if kpoint_gamma else [1, 1, 2],
        symmetry='off',
        convergence={'maximum iterations': 3})
    return calculator


@pytest.mark.parametrize('kpoint_gamma', [True, False])
def test_davidson_scalapack_eigenvalues(scalapack, kpoint_gamma):
    atoms = bulk('Si', cubic=True) * [2, 2, 2]

    atoms1 = atoms.copy()
    atoms2 = atoms.copy()

    atoms1.calc = get_calculator(sl_auto=True, kpoint_gamma=kpoint_gamma)
    atoms2.calc = get_calculator(sl_auto=False, kpoint_gamma=kpoint_gamma)

    e1 = atoms1.get_potential_energy()
    e2 = atoms2.get_potential_energy()

    for n, kpoint in enumerate(atoms1.calc.get_bz_k_points()):
        eigvals1 = atoms1.calc.get_eigenvalues(kpt=n, spin=0)
        eigvals2 = atoms2.calc.get_eigenvalues(kpt=n, spin=0)

        assert np.allclose(eigvals1, eigvals2, rtol=1e-11)

    assert e1 == pytest.approx(e2, rel=1e-11)



import pytest
import numpy as np
import time

from ase.build import bulk
from ase.parallel import parprint

from gpaw import GPAW, PW
from gpaw.test import findpeak, equal
from gpaw.mpi import size, world

from gpaw.response import ResponseGroundStateAdapter
from gpaw.response.chiks import ChiKSCalculator
from gpaw.response.susceptibility import ChiFactory
from gpaw.response.pair_functions import read_susceptibility_array


@pytest.mark.kspair
@pytest.mark.response
def test_response_two_aluminum_chi_RPA(in_tmp_dir):
    assert size <= 4**3

    # Ground state calculation

    t1 = time.time()

    a = 4.043
    atoms1 = bulk('Al', 'fcc', a=a)
    atoms2 = atoms1.repeat((2, 1, 1))

    calc1 = GPAW(mode=PW(200),
                 nbands=4,
                 kpts=(8, 8, 8),
                 parallel={'domain': 1},
                 xc='LDA')

    atoms1.calc = calc1
    atoms1.get_potential_energy()

    t2 = time.time()

    calc2 = GPAW(mode=PW(200),
                 nbands=8,
                 kpts=(4, 8, 8),
                 parallel={'domain': 1},
                 xc='LDA')

    atoms2.calc = calc2
    atoms2.get_potential_energy()

    t3 = time.time()

    # Excited state calculation
    q1_qc = [np.array([1 / 8., 0., 0.]), np.array([3 / 8., 0., 0.])]
    q2_qc = [np.array([1 / 4., 0., 0.]), np.array([- 1 / 4., 0., 0.])]
    w = np.linspace(0, 24, 241)

    # Calculate susceptibility using Al1
    calculate_chi(calc1, q1_qc, w, filename_prefix='Al1')

    t4 = time.time()

    # Calculate susceptibility using Al2
    calculate_chi(calc2, q2_qc, w, filename_prefix='Al2')

    t5 = time.time()

    parprint('')
    parprint('Ground state calc 1 took', (t2 - t1), 'seconds')
    parprint('Ground state calc 2 took', (t3 - t2), 'seconds')
    parprint('Susceptibility calc 1 took', (t4 - t3), 'seconds')
    parprint('Susceptibility calc 2 took', (t5 - t4), 'seconds')

    # Check that results are consistent, when structure is simply repeated

    # Read results
    w11_w, G11_Gc, chi11_wGG = read_susceptibility_array('Al1_chiGG_q1.pckl')
    w21_w, G21_Gc, chi21_wGG = read_susceptibility_array('Al2_chiGG_q1.pckl')
    w12_w, G12_Gc, chi12_wGG = read_susceptibility_array('Al1_chiGG_q2.pckl')
    w22_w, G22_Gc, chi22_wGG = read_susceptibility_array('Al2_chiGG_q2.pckl')

    # Check that reciprocal lattice vectors remain as assumed in check below
    equal(np.linalg.norm(G11_Gc[0]), 0., 1e-6)
    equal(np.linalg.norm(G21_Gc[0]), 0., 1e-6)
    equal(np.linalg.norm(G12_Gc[0]), 0., 1e-6)
    equal(np.linalg.norm(G22_Gc[1] - np.array([1, 0, 0])), 0., 1e-6)

    # Check plasmon peaks remain the same
    wpeak11, Ipeak11 = findpeak(w11_w, -chi11_wGG[:, 0, 0].imag)
    wpeak21, Ipeak21 = findpeak(w21_w, -chi21_wGG[:, 0, 0].imag)
    equal(wpeak11, wpeak21, 0.02)
    equal(Ipeak11, Ipeak21, 1.0)
    wpeak12, Ipeak12 = findpeak(w12_w, -chi12_wGG[:, 0, 0].imag)
    wpeak22, Ipeak22 = findpeak(w22_w, -chi22_wGG[:, 1, 1].imag)
    equal(wpeak12, wpeak22, 0.05)
    equal(Ipeak12, Ipeak22, 1.0)


def calculate_chi(calc, q_qc, w,
                  eta=0.2, ecut=50,
                  spincomponent='00', fxc='RPA',
                  filename_prefix=None, reduced_ecut=25):
    gs = ResponseGroundStateAdapter(calc)
    chiks_calc = ChiKSCalculator(gs, ecut=ecut)
    chi_factory = ChiFactory(chiks_calc)

    if filename_prefix is None:
        filename = 'chiGG_qXXX.pckl'
    else:
        filename = filename_prefix + '_chiGG_qXXX.pckl'

    for q, q_c in enumerate(q_qc):
        fname = filename.replace('XXX', str(q + 1))
        _, chi = chi_factory(spincomponent, q_c, w + 1.j * eta, fxc=fxc)
        chi = chi.copy_with_reduced_ecut(reduced_ecut)
        chi.write_array(fname)
        world.barrier()
